import logging
import os
import sys
from abc import ABC, abstractmethod
from contextlib import ExitStack, contextmanager
from typing import Callable, Dict, Iterator, List, Optional, Sequence, Tuple, Union

import matplotlib
import numpy as np
from tqdm.auto import tqdm
from typing_extensions import TypedDict

from qcodes import config
from qcodes.dataset.data_set import res_type
from qcodes.dataset.data_set_protocol import DataSetProtocol
from qcodes.dataset.descriptions.detect_shapes import detect_shape_of_measurement
from qcodes.dataset.descriptions.versioning.rundescribertypes import Shapes
from qcodes.dataset.experiment_container import Experiment
from qcodes.dataset.measurements import Measurement
from qcodes.dataset.plotting import plot_dataset
from qcodes.instrument.parameter import _BaseParameter
from qcodes.utils.threading import (
    SequentialParamsCaller,
    ThreadPoolParamsCaller,
    process_params_meas,
)

ActionsT = Sequence[Callable[[], None]]

ParamMeasT = Union[_BaseParameter, Callable[[], None]]

AxesTuple = Tuple[matplotlib.axes.Axes, matplotlib.colorbar.Colorbar]
AxesTupleList = Tuple[
    List[matplotlib.axes.Axes], List[Optional[matplotlib.colorbar.Colorbar]]
]
AxesTupleListWithDataSet = Tuple[
    DataSetProtocol,
    List[matplotlib.axes.Axes],
    List[Optional[matplotlib.colorbar.Colorbar]],
]
MultiAxesTupleListWithDataSet = Tuple[
    Tuple[DataSetProtocol, ...],
    Tuple[List[matplotlib.axes.Axes], ...],
    Tuple[List[Optional[matplotlib.colorbar.Colorbar]], ...],
]

LOG = logging.getLogger(__name__)


class ParameterGroup(TypedDict):
    params: Tuple[ParamMeasT, ...]
    meas_name: str
    measured_params: List[res_type]


class UnsafeThreadingException(Exception):
    pass


def _register_parameters(
    meas: Measurement,
    param_meas: Sequence[ParamMeasT],
    setpoints: Optional[Sequence[_BaseParameter]] = None,
    shapes: Shapes = None,
) -> None:
    for parameter in param_meas:
        if isinstance(parameter, _BaseParameter):
            meas.register_parameter(parameter, setpoints=setpoints)
    meas.set_shapes(shapes=shapes)


def _register_actions(
    meas: Measurement, enter_actions: ActionsT, exit_actions: ActionsT
) -> None:
    for action in enter_actions:
        # this omits the possibility of passing
        # argument to enter and exit actions.
        # Do we want that?
        meas.add_before_run(action, ())
    for action in exit_actions:
        meas.add_after_run(action, ())


def _set_write_period(meas: Measurement, write_period: Optional[float] = None) -> None:
    if write_period is not None:
        meas.write_period = write_period


@contextmanager
def _catch_keyboard_interrupts() -> Iterator[Callable[[], bool]]:
    interrupted = False

    def has_been_interrupted() -> bool:
        nonlocal interrupted
        return interrupted

    try:
        yield has_been_interrupted
    except KeyboardInterrupt:
        interrupted = True


def do0d(
    *param_meas: ParamMeasT,
    write_period: Optional[float] = None,
    measurement_name: str = "",
    exp: Optional[Experiment] = None,
    do_plot: Optional[bool] = None,
    use_threads: Optional[bool] = None,
    log_info: Optional[str] = None,
) -> AxesTupleListWithDataSet:
    """
    Perform a measurement of a single parameter. This is probably most
    useful for an ArrayParameter that already returns an array of data points

    Args:
        *param_meas: Parameter(s) to measure at each step or functions that
          will be called at each step. The function should take no arguments.
          The parameters and functions are called in the order they are
          supplied.
        write_period: The time after which the data is actually written to the
            database.
        measurement_name: Name of the measurement. This will be passed down to
            the dataset produced by the measurement. If not given, a default
            value of 'results' is used for the dataset.
        exp: The experiment to use for this measurement.
        do_plot: should png and pdf versions of the images be saved after the
            run. If None the setting will be read from ``qcodesrc.json``
        use_threads: If True measurements from each instrument will be done on
            separate threads. If you are measuring from several instruments
            this may give a significant speedup.
        log_info: Message that is logged during the measurement. If None a default
            message is used.

    Returns:
        The QCoDeS dataset.
    """
    if do_plot is None:
        do_plot = config.dataset.dond_plot
    meas = Measurement(name=measurement_name, exp=exp)
    if log_info is not None:
        meas._extra_log_info = log_info
    else:
        meas._extra_log_info = "Using 'qcodes.utils.dataset.doNd.do0d'"

    measured_parameters = tuple(
        param for param in param_meas if isinstance(param, _BaseParameter)
    )

    try:
        shapes: Shapes = detect_shape_of_measurement(
            measured_parameters,
        )
    except TypeError:
        LOG.exception(
            f"Could not detect shape of {measured_parameters} "
            f"falling back to unknown shape."
        )
        shapes = None

    _register_parameters(meas, param_meas, shapes=shapes)
    _set_write_period(meas, write_period)

    with meas.run() as datasaver:
        datasaver.add_result(*process_params_meas(param_meas, use_threads=use_threads))
        dataset = datasaver.dataset

    return _handle_plotting(dataset, do_plot)


def do1d(
    param_set: _BaseParameter,
    start: float,
    stop: float,
    num_points: int,
    delay: float,
    *param_meas: ParamMeasT,
    enter_actions: ActionsT = (),
    exit_actions: ActionsT = (),
    write_period: Optional[float] = None,
    measurement_name: str = "",
    exp: Optional[Experiment] = None,
    do_plot: Optional[bool] = None,
    use_threads: Optional[bool] = None,
    additional_setpoints: Sequence[ParamMeasT] = tuple(),
    show_progress: Optional[None] = None,
    log_info: Optional[str] = None,
) -> AxesTupleListWithDataSet:
    """
    Perform a 1D scan of ``param_set`` from ``start`` to ``stop`` in
    ``num_points`` measuring param_meas at each step. In case param_meas is
    an ArrayParameter this is effectively a 2d scan.

    Args:
        param_set: The QCoDeS parameter to sweep over
        start: Starting point of sweep
        stop: End point of sweep
        num_points: Number of points in sweep
        delay: Delay after setting parameter before measurement is performed
        *param_meas: Parameter(s) to measure at each step or functions that
          will be called at each step. The function should take no arguments.
          The parameters and functions are called in the order they are
          supplied.
        enter_actions: A list of functions taking no arguments that will be
            called before the measurements start
        exit_actions: A list of functions taking no arguments that will be
            called after the measurements ends
        write_period: The time after which the data is actually written to the
            database.
        additional_setpoints: A list of setpoint parameters to be registered in
            the measurement but not scanned.
        measurement_name: Name of the measurement. This will be passed down to
            the dataset produced by the measurement. If not given, a default
            value of 'results' is used for the dataset.
        exp: The experiment to use for this measurement.
        do_plot: should png and pdf versions of the images be saved after the
            run. If None the setting will be read from ``qcodesrc.json`
        use_threads: If True measurements from each instrument will be done on
            separate threads. If you are measuring from several instruments
            this may give a significant speedup.
        show_progress: should a progress bar be displayed during the
            measurement. If None the setting will be read from ``qcodesrc.json`
        log_info: Message that is logged during the measurement. If None a default
            message is used.

    Returns:
        The QCoDeS dataset.
    """
    if do_plot is None:
        do_plot = config.dataset.dond_plot
    if show_progress is None:
        show_progress = config.dataset.dond_show_progress

    meas = Measurement(name=measurement_name, exp=exp)
    if log_info is not None:
        meas._extra_log_info = log_info
    else:
        meas._extra_log_info = "Using 'qcodes.utils.dataset.doNd.do1d'"

    all_setpoint_params = (param_set,) + tuple(s for s in additional_setpoints)

    measured_parameters = tuple(
        param for param in param_meas if isinstance(param, _BaseParameter)
    )
    try:
        loop_shape = (num_points,) + tuple(1 for _ in additional_setpoints)
        shapes: Shapes = detect_shape_of_measurement(measured_parameters, loop_shape)
    except TypeError:
        LOG.exception(
            f"Could not detect shape of {measured_parameters} "
            f"falling back to unknown shape."
        )
        shapes = None

    _register_parameters(meas, all_setpoint_params)
    _register_parameters(meas, param_meas, setpoints=all_setpoint_params, shapes=shapes)
    _set_write_period(meas, write_period)
    _register_actions(meas, enter_actions, exit_actions)

    original_delay = param_set.post_delay
    param_set.post_delay = delay

    if use_threads is None:
        use_threads = config.dataset.use_threads

    param_meas_caller = (
        ThreadPoolParamsCaller(*param_meas)
        if use_threads
        else SequentialParamsCaller(*param_meas)
    )

    # do1D enforces a simple relationship between measured parameters
    # and set parameters. For anything more complicated this should be
    # reimplemented from scratch
    with _catch_keyboard_interrupts() as interrupted, meas.run() as datasaver, param_meas_caller as call_param_meas:
        dataset = datasaver.dataset
        additional_setpoints_data = process_params_meas(additional_setpoints)
        setpoints = np.linspace(start, stop, num_points)

        # flush to prevent unflushed print's to visually interrupt tqdm bar
        # updates
        sys.stdout.flush()
        sys.stderr.flush()

        for set_point in tqdm(setpoints, disable=not show_progress):
            param_set.set(set_point)
            datasaver.add_result(
                (param_set, set_point), *call_param_meas(), *additional_setpoints_data
            )

    param_set.post_delay = original_delay

    return _handle_plotting(dataset, do_plot, interrupted())


def do2d(
    param_set1: _BaseParameter,
    start1: float,
    stop1: float,
    num_points1: int,
    delay1: float,
    param_set2: _BaseParameter,
    start2: float,
    stop2: float,
    num_points2: int,
    delay2: float,
    *param_meas: ParamMeasT,
    set_before_sweep: Optional[bool] = True,
    enter_actions: ActionsT = (),
    exit_actions: ActionsT = (),
    before_inner_actions: ActionsT = (),
    after_inner_actions: ActionsT = (),
    write_period: Optional[float] = None,
    measurement_name: str = "",
    exp: Optional[Experiment] = None,
    flush_columns: bool = False,
    do_plot: Optional[bool] = None,
    use_threads: Optional[bool] = None,
    additional_setpoints: Sequence[ParamMeasT] = tuple(),
    show_progress: Optional[None] = None,
    log_info: Optional[str] = None,
) -> AxesTupleListWithDataSet:
    """
    Perform a 1D scan of ``param_set1`` from ``start1`` to ``stop1`` in
    ``num_points1`` and ``param_set2`` from ``start2`` to ``stop2`` in
    ``num_points2`` measuring param_meas at each step.

    Args:
        param_set1: The QCoDeS parameter to sweep over in the outer loop
        start1: Starting point of sweep in outer loop
        stop1: End point of sweep in the outer loop
        num_points1: Number of points to measure in the outer loop
        delay1: Delay after setting parameter in the outer loop
        param_set2: The QCoDeS parameter to sweep over in the inner loop
        start2: Starting point of sweep in inner loop
        stop2: End point of sweep in the inner loop
        num_points2: Number of points to measure in the inner loop
        delay2: Delay after setting parameter before measurement is performed
        *param_meas: Parameter(s) to measure at each step or functions that
          will be called at each step. The function should take no arguments.
          The parameters and functions are called in the order they are
          supplied.
        set_before_sweep: if True the outer parameter is set to its first value
            before the inner parameter is swept to its next value.
        enter_actions: A list of functions taking no arguments that will be
            called before the measurements start
        exit_actions: A list of functions taking no arguments that will be
            called after the measurements ends
        before_inner_actions: Actions executed before each run of the inner loop
        after_inner_actions: Actions executed after each run of the inner loop
        write_period: The time after which the data is actually written to the
            database.
        measurement_name: Name of the measurement. This will be passed down to
            the dataset produced by the measurement. If not given, a default
            value of 'results' is used for the dataset.
        exp: The experiment to use for this measurement.
        flush_columns: The data is written after a column is finished
            independent of the passed time and write period.
        additional_setpoints: A list of setpoint parameters to be registered in
            the measurement but not scanned.
        do_plot: should png and pdf versions of the images be saved after the
            run. If None the setting will be read from ``qcodesrc.json``
        use_threads: If True measurements from each instrument will be done on
            separate threads. If you are measuring from several instruments
            this may give a significant speedup.
        show_progress: should a progress bar be displayed during the
            measurement. If None the setting will be read from ``qcodesrc.json`
        log_info: Message that is logged during the measurement. If None a default
            message is used.

    Returns:
        The QCoDeS dataset.
    """

    if do_plot is None:
        do_plot = config.dataset.dond_plot
    if show_progress is None:
        show_progress = config.dataset.dond_show_progress

    meas = Measurement(name=measurement_name, exp=exp)
    if log_info is not None:
        meas._extra_log_info = log_info
    else:
        meas._extra_log_info = "Using 'qcodes.utils.dataset.doNd.do2d'"
    all_setpoint_params = (
        param_set1,
        param_set2,
    ) + tuple(s for s in additional_setpoints)

    measured_parameters = tuple(
        param for param in param_meas if isinstance(param, _BaseParameter)
    )

    try:
        loop_shape = (num_points1, num_points2) + tuple(1 for _ in additional_setpoints)
        shapes: Shapes = detect_shape_of_measurement(measured_parameters, loop_shape)
    except TypeError:
        LOG.exception(
            f"Could not detect shape of {measured_parameters} "
            f"falling back to unknown shape."
        )
        shapes = None

    _register_parameters(meas, all_setpoint_params)
    _register_parameters(meas, param_meas, setpoints=all_setpoint_params, shapes=shapes)
    _set_write_period(meas, write_period)
    _register_actions(meas, enter_actions, exit_actions)

    original_delay1 = param_set1.post_delay
    original_delay2 = param_set2.post_delay

    param_set1.post_delay = delay1
    param_set2.post_delay = delay2

    if use_threads is None:
        use_threads = config.dataset.use_threads

    param_meas_caller = (
        ThreadPoolParamsCaller(*param_meas)
        if use_threads
        else SequentialParamsCaller(*param_meas)
    )

    with _catch_keyboard_interrupts() as interrupted, meas.run() as datasaver, param_meas_caller as call_param_meas:
        dataset = datasaver.dataset
        additional_setpoints_data = process_params_meas(additional_setpoints)
        setpoints1 = np.linspace(start1, stop1, num_points1)
        for set_point1 in tqdm(setpoints1, disable=not show_progress):
            if set_before_sweep:
                param_set2.set(start2)

            param_set1.set(set_point1)

            for action in before_inner_actions:
                action()

            setpoints2 = np.linspace(start2, stop2, num_points2)

            # flush to prevent unflushed print's to visually interrupt tqdm bar
            # updates
            sys.stdout.flush()
            sys.stderr.flush()
            for set_point2 in tqdm(setpoints2, disable=not show_progress, leave=False):
                # skip first inner set point if `set_before_sweep`
                if set_point2 == start2 and set_before_sweep:
                    pass
                else:
                    param_set2.set(set_point2)

                datasaver.add_result(
                    (param_set1, set_point1),
                    (param_set2, set_point2),
                    *call_param_meas(),
                    *additional_setpoints_data,
                )

            for action in after_inner_actions:
                action()
            if flush_columns:
                datasaver.flush_data_to_database()

    param_set1.post_delay = original_delay1
    param_set2.post_delay = original_delay2

    return _handle_plotting(dataset, do_plot, interrupted())


class AbstractSweep(ABC):
    """
    Abstract sweep class that defines an interface for concrete sweep classes.
    """

    @abstractmethod
    def get_setpoints(self) -> np.ndarray:
        """
        Returns an array of setpoint values for this sweep.
        """
        pass

    @property
    @abstractmethod
    def param(self) -> _BaseParameter:
        """
        Returns the Qcodes sweep parameter.
        """
        pass

    @property
    @abstractmethod
    def delay(self) -> float:
        """
        Delay between two consecutive sweep points.
        """
        pass

    @property
    @abstractmethod
    def num_points(self) -> int:
        """
        Number of sweep points.
        """
        pass

    @property
    @abstractmethod
    def post_actions(self) -> ActionsT:
        """
        actions to be performed after setting param to its setpoint.
        """
        pass


class LinSweep(AbstractSweep):
    """
    Linear sweep.

    Args:
        param: Qcodes parameter to sweep.
        start: Sweep start value.
        stop: Sweep end value.
        num_points: Number of sweep points.
        delay: Time in seconds between two consequtive sweep points
    """

    def __init__(
        self,
        param: _BaseParameter,
        start: float,
        stop: float,
        num_points: int,
        delay: float = 0,
        post_actions: ActionsT = (),
    ):
        self._param = param
        self._start = start
        self._stop = stop
        self._num_points = num_points
        self._delay = delay
        self._post_actions = post_actions

    def get_setpoints(self) -> np.ndarray:
        """
        Linear (evenly spaced) numpy array for supplied start, stop and
        num_points.
        """
        return np.linspace(self._start, self._stop, self._num_points)

    @property
    def param(self) -> _BaseParameter:
        return self._param

    @property
    def delay(self) -> float:
        return self._delay

    @property
    def num_points(self) -> int:
        return self._num_points

    @property
    def post_actions(self) -> ActionsT:
        return self._post_actions


class LogSweep(AbstractSweep):
    """
    Logarithmic sweep.

    Args:
        param: Qcodes parameter for sweep.
        start: Sweep start value.
        stop: Sweep end value.
        num_points: Number of sweep points.
        delay: Time in seconds between two consequtive sweep points.
    """

    def __init__(
        self,
        param: _BaseParameter,
        start: float,
        stop: float,
        num_points: int,
        delay: float = 0,
        post_actions: ActionsT = (),
    ):
        self._param = param
        self._start = start
        self._stop = stop
        self._num_points = num_points
        self._delay = delay
        self._post_actions = post_actions

    def get_setpoints(self) -> np.ndarray:
        """
        Logarithmically spaced numpy array for supplied start, stop and
        num_points.
        """
        return np.logspace(self._start, self._stop, self._num_points)

    @property
    def param(self) -> _BaseParameter:
        return self._param

    @property
    def delay(self) -> float:
        return self._delay

    @property
    def num_points(self) -> int:
        return self._num_points

    @property
    def post_actions(self) -> ActionsT:
        return self._post_actions


def dond(
    *params: Union[AbstractSweep, Union[ParamMeasT, Sequence[ParamMeasT]]],
    write_period: Optional[float] = None,
    measurement_name: str = "",
    exp: Optional[Experiment] = None,
    enter_actions: ActionsT = (),
    exit_actions: ActionsT = (),
    do_plot: Optional[bool] = None,
    show_progress: Optional[bool] = None,
    use_threads: Optional[bool] = None,
    additional_setpoints: Sequence[ParamMeasT] = tuple(),
    log_info: Optional[str] = None,
) -> Union[AxesTupleListWithDataSet, MultiAxesTupleListWithDataSet]:
    """
    Perform n-dimentional scan from slowest (first) to the fastest (last), to
    measure m measurement parameters. The dimensions should be specified
    as sweep objects, and after them the parameters to measure should be passed.

    Args:
        *params: Instances of n sweep classes and m measurement parameters,
            e.g. if linear sweep is considered:

            .. code-block::

                LinSweep(param_set_1, start_1, stop_1, num_points_1, delay_1), ...,
                LinSweep(param_set_n, start_n, stop_n, num_points_n, delay_n),
                param_meas_1, param_meas_2, ..., param_meas_m

            If multiple DataSets creation is needed, measurement parameters should
            be grouped, so one dataset will be created for each group. e.g.:

            .. code-block::

                LinSweep(param_set_1, start_1, stop_1, num_points_1, delay_1), ...,
                LinSweep(param_set_n, start_n, stop_n, num_points_n, delay_n),
                [param_meas_1, param_meas_2], ..., [param_meas_m]

        write_period: The time after which the data is actually written to the
            database.
        measurement_name: Name of the measurement. This will be passed down to
            the dataset produced by the measurement. If not given, a default
            value of 'results' is used for the dataset.
        exp: The experiment to use for this measurement.
        enter_actions: A list of functions taking no arguments that will be
            called before the measurements start.
        exit_actions: A list of functions taking no arguments that will be
            called after the measurements ends.
        do_plot: should png and pdf versions of the images be saved and plots
            are shown after the run. If None the setting will be read from
            ``qcodesrc.json``
        show_progress: should a progress bar be displayed during the
            measurement. If None the setting will be read from ``qcodesrc.json`
        use_threads: If True, measurements from each instrument will be done on
            separate threads. If you are measuring from several instruments
            this may give a significant speedup.
        additional_setpoints: A list of setpoint parameters to be registered in
            the measurement but not scanned/swept-over.
        log_info: Message that is logged during the measurement. If None a default
            message is used.

        Returns:
            A tuple of QCoDeS DataSet, Matplotlib axis, Matplotlib colorbar. If
            more than one group of measurement parameters is supplied, the output
            will be a tuple of tuple(QCoDeS DataSet), tuple(Matplotlib axis),
            tuple(Matplotlib colorbar), in which each element of each sub-tuple
            belongs to one group, and the order of elements is the order of
            the supplied groups.
    """
    if do_plot is None:
        do_plot = config.dataset.dond_plot
    if show_progress is None:
        show_progress = config.dataset.dond_show_progress

    sweep_instances, params_meas = _parse_dond_arguments(*params)
    nested_setpoints = _make_nested_setpoints(sweep_instances)

    all_setpoint_params = tuple(sweep.param for sweep in sweep_instances) + tuple(
        s for s in additional_setpoints
    )

    (
        all_meas_parameters,
        grouped_parameters,
        measured_parameters,
    ) = _extract_paramters_by_type_and_group(measurement_name, params_meas)

    try:
        loop_shape = tuple(sweep.num_points for sweep in sweep_instances) + tuple(
            1 for _ in additional_setpoints
        )
        shapes: Shapes = detect_shape_of_measurement(measured_parameters, loop_shape)
    except TypeError:
        LOG.exception(
            f"Could not detect shape of {measured_parameters} "
            f"falling back to unknown shape."
        )
        shapes = None
    meas_list = _create_measurements(
        all_setpoint_params,
        enter_actions,
        exit_actions,
        exp,
        grouped_parameters,
        shapes,
        write_period,
        log_info,
    )

    original_delays: Dict[_BaseParameter, float] = {}
    params_set: List[_BaseParameter] = []
    post_actions: List[ActionsT] = []
    for sweep in sweep_instances:
        original_delays[sweep.param] = sweep.param.post_delay
        sweep.param.post_delay = sweep.delay
        params_set.append(sweep.param)
        post_actions.append(sweep.post_actions)

    datasets = []
    plots_axes = []
    plots_colorbar = []
    if use_threads is None:
        use_threads = config.dataset.use_threads

    params_meas_caller = (
        ThreadPoolParamsCaller(*all_meas_parameters)
        if use_threads
        else SequentialParamsCaller(*all_meas_parameters)
    )

    try:
        with _catch_keyboard_interrupts() as interrupted, ExitStack() as stack, params_meas_caller as call_params_meas:
            datasavers = [stack.enter_context(measure.run()) for measure in meas_list]
            additional_setpoints_data = process_params_meas(additional_setpoints)
            previous_setpoints = np.empty(len(sweep_instances))
            for setpoints in tqdm(nested_setpoints, disable=not show_progress):

                active_actions = _select_active_actions(
                    post_actions, setpoints, previous_setpoints
                )
                previous_setpoints = setpoints

                param_set_list = []
                param_value_action = zip(params_set, setpoints, active_actions)
                for setpoint_param, setpoint, action in param_value_action:
                    setpoint_param(setpoint)
                    param_set_list.append((setpoint_param, setpoint))
                    for act in action:
                        act()

                meas_value_pair = call_params_meas()
                for group in grouped_parameters.values():
                    group["measured_params"] = []
                    for measured in meas_value_pair:
                        if measured[0] in group["params"]:
                            group["measured_params"].append(measured)
                for ind, datasaver in enumerate(datasavers):
                    datasaver.add_result(
                        *param_set_list,
                        *grouped_parameters[f"group_{ind}"]["measured_params"],
                        *additional_setpoints_data,
                    )

    finally:
        for parameter, original_delay in original_delays.items():
            parameter.post_delay = original_delay

        for datasaver in datasavers:
            ds, plot_axis, plot_color = _handle_plotting(
                datasaver.dataset, do_plot, interrupted()
            )
            datasets.append(ds)
            plots_axes.append(plot_axis)
            plots_colorbar.append(plot_color)

    if len(grouped_parameters) == 1:
        return datasets[0], plots_axes[0], plots_colorbar[0]
    else:
        return tuple(datasets), tuple(plots_axes), tuple(plots_colorbar)


def _parse_dond_arguments(
        *params: Union[AbstractSweep, Union[ParamMeasT, Sequence[ParamMeasT]]]
    ) -> Tuple[List[AbstractSweep], List[Union[ParamMeasT, Sequence[ParamMeasT]]]]:
        """
        Parse supplied arguments into sweep objects and measurement parameters
        and their callables.
        """
        sweep_instances: List[AbstractSweep] = []
        params_meas: List[Union[ParamMeasT, Sequence[ParamMeasT]]] = []
        for par in params:
            if isinstance(par, AbstractSweep):
                sweep_instances.append(par)
            else:
                params_meas.append(par)
        return sweep_instances, params_meas


def _make_nested_setpoints(sweeps: List[AbstractSweep]) -> np.ndarray:
        """Create the cartesian product of all the setpoint values."""
        if len(sweeps) == 0:
            return np.array([[]])  # 0d sweep (do0d)
        setpoint_values = [sweep.get_setpoints() for sweep in sweeps]
        setpoint_grids = np.meshgrid(*setpoint_values, indexing="ij")
        flat_setpoint_grids = [np.ravel(grid, order="C") for grid in setpoint_grids]
        return np.vstack(flat_setpoint_grids).T


def _select_active_actions(
    actions: Sequence[ActionsT], setpoints: np.ndarray, previous_setpoints: np.ndarray
) -> List[ActionsT]:
    """
    Select ActionT (Sequence[Callable]) from a Sequence of ActionsT if
    the corresponding setpoint has changed. Otherwise select an empty Sequence.
    """
    actions_list: List[ActionsT] = [()] * len(setpoints)
    for ind, (new_setpoint, old_setpoint) in enumerate(
        zip(setpoints, previous_setpoints)
    ):
        if new_setpoint != old_setpoint:
            actions_list[ind] = actions[ind]
    return actions_list


def _create_measurements(
    all_setpoint_params: Sequence[_BaseParameter],
    enter_actions: ActionsT,
    exit_actions: ActionsT,
    exp: Optional[Experiment],
    grouped_parameters: Dict[str, ParameterGroup],
    shapes: Shapes,
    write_period: Optional[float],
    log_info: Optional[str],
) -> Tuple[Measurement, ...]:
    meas_list: List[Measurement] = []
    if log_info is not None:
        _extra_log_info = log_info
    else:
        _extra_log_info = "Using 'qcodes.utils.dataset.doNd.dond'"
    for group in grouped_parameters.values():
        meas_name = group["meas_name"]
        meas_params = group["params"]
        meas = Measurement(name=meas_name, exp=exp)
        meas._extra_log_info = _extra_log_info
        _register_parameters(meas, all_setpoint_params)
        _register_parameters(
            meas, meas_params, setpoints=all_setpoint_params, shapes=shapes
        )
        _set_write_period(meas, write_period)
        _register_actions(meas, enter_actions, exit_actions)
        meas_list.append(meas)
    return tuple(meas_list)


def _extract_paramters_by_type_and_group(
    measurement_name: str,
    params_meas: Sequence[Union[ParamMeasT, Sequence[ParamMeasT]]],
) -> Tuple[
    Tuple[ParamMeasT, ...], Dict[str, ParameterGroup], Tuple[_BaseParameter, ...]
]:
    measured_parameters: List[_BaseParameter] = []
    all_meas_parameters: List[ParamMeasT] = []
    single_group: List[ParamMeasT] = []
    multi_group: List[Sequence[ParamMeasT]] = []
    grouped_parameters: Dict[str, ParameterGroup] = {}
    for param in params_meas:
        if not isinstance(param, Sequence):
            single_group.append(param)
            all_meas_parameters.append(param)
            if isinstance(param, _BaseParameter):
                measured_parameters.append(param)
        elif not isinstance(param, str):
            multi_group.append(param)
            for nested_param in param:
                all_meas_parameters.append(nested_param)
                if isinstance(nested_param, _BaseParameter):
                    measured_parameters.append(nested_param)
    if single_group:
        pg: ParameterGroup = {
            "params": tuple(single_group),
            "meas_name": measurement_name,
            "measured_params": [],
        }
        grouped_parameters["group_0"] = pg
    if multi_group:
        for index, par in enumerate(multi_group):
            pg = {
                "params": tuple(par),
                "meas_name": measurement_name,
                "measured_params": [],
            }
            grouped_parameters[f"group_{index}"] = pg
    return tuple(all_meas_parameters), grouped_parameters, tuple(measured_parameters)


def _handle_plotting(
    data: DataSetProtocol, do_plot: bool = True, interrupted: bool = False
) -> AxesTupleListWithDataSet:
    """
    Save the plots created by datasaver as pdf and png

    Args:
        datasaver: a measurement datasaver that contains a dataset to be saved
            as plot.
            :param do_plot:

    """
    if do_plot:
        res = plot(data)
    else:
        res = data, [None], [None]

    if interrupted:
        raise KeyboardInterrupt

    return res


def plot(
    data: DataSetProtocol, save_pdf: bool = True, save_png: bool = True
) -> Tuple[
    DataSetProtocol,
    List[matplotlib.axes.Axes],
    List[Optional[matplotlib.colorbar.Colorbar]],
]:
    """
    The utility function to plot results and save the figures either in pdf or
    png or both formats.

    Args:
        data: The QCoDeS dataset to be plotted.
        save_pdf: Save figure in pdf format.
        save_png: Save figure in png format.
    """
    dataid = data.captured_run_id
    axes, cbs = plot_dataset(data)
    mainfolder = config.user.mainfolder
    experiment_name = data.exp_name
    sample_name = data.sample_name
    storage_dir = os.path.join(mainfolder, experiment_name, sample_name)
    os.makedirs(storage_dir, exist_ok=True)
    png_dir = os.path.join(storage_dir, "png")
    pdf_dif = os.path.join(storage_dir, "pdf")
    os.makedirs(png_dir, exist_ok=True)
    os.makedirs(pdf_dif, exist_ok=True)
    for i, ax in enumerate(axes):
        if save_pdf:
            full_path = os.path.join(pdf_dif, f"{dataid}_{i}.pdf")
            ax.figure.savefig(full_path, dpi=500)
        if save_png:
            full_path = os.path.join(png_dir, f"{dataid}_{i}.png")
            ax.figure.savefig(full_path, dpi=500)
    res = data, axes, cbs
    return res



import pytest

from qcodes.tests.common import error_caused_by


def test_get_metadata_from_dataset(dataset):
    dataset.add_metadata('something', 123)
    something = dataset.get_metadata('something')
    assert 123 == something


def test_get_nonexisting_metadata(dataset):

    data = dataset.get_metadata("something")
    assert data is None



"""
A helper module containing a class to keep track of vectors in different
coordinate systems.
"""


from typing import Any, List, Optional, Sequence, Type, TypeVar, Union

import numpy as np

NormOrder = Union[str, float]
T = TypeVar('T', bound='FieldVector')


class FieldVector:
    """
    A convenient class to keep track of vectors representing physical fields.
    The idea is that a vector instance stores a representation in Cartesian,
    spherical and cylindrical coordinates. All arguments are optional, however
    the user needs to provide one of the combinations of either (x, y, z) values
    or (rho, phi, z) values or (r, theta, phi) values at instantiation for a
    meaningful computation of the other representation, immediately.
    """
    attributes = ["x", "y", "z", "r", "theta", "phi", "rho"]
    repr_format = "cartesian"

    def __init__(self,
                 x: Optional[float] = None,
                 y: Optional[float] = None,
                 z: Optional[float] = None,
                 r: Optional[float] = None,
                 theta: Optional[float] = None,
                 phi: Optional[float] = None,
                 rho: Optional[float] = None):
        """
        Args:
            x: represents the norm of the projection
                of the vector along the x-axis
            y: represents the norm of the projection
                of the vector along the y-axis
            z: represents the norm of the projection
                of the vector along the z-axis
            r: represents the norm of the vector
            theta: represents the angle of the vector
                with respect to the positive z-axis
            rho: represents the norm of the projection
                of the vector on to the xy-plane
            phi: represents the angle of rho
                with respect to the positive x-axis
        """

        self._x = float(x) if x is not None else None
        self._y = float(y) if y is not None else None
        self._z = float(z) if z is not None else None

        self._r = float(r) if r is not None else None
        self._theta = float(np.radians(theta)) if theta is not None else None
        self._phi = float(np.radians(phi)) if phi is not None else None
        self._rho = float(rho) if rho is not None else None

        self._compute_unknowns()

    def _set_attribute_value(self, attr_name: str, value: Optional[float]) -> None:
        if value is None:
            return

        attr_value = getattr(self, "_" + attr_name)

        if attr_value is None:
            setattr(self, "_" + attr_name, value)
        else:
            if not np.isclose(attr_value, value):
                raise ValueError(
                    f"Computed value of {attr_name} inconsistent with given "
                    f"value"
                )

    def _set_attribute_values(
        self, attr_names: Sequence[str], values: Sequence[Optional[float]]
    ) -> None:
        for attr_name, value in zip(attr_names, values):
            self._set_attribute_value(attr_name, value)

    def __getnewargs__(self):
        return self.x, self.y, self.z

    @staticmethod
    def _cartesian_to_other(x, y, z):
        """Convert a cartesian set of coordinates to values of interest."""
        if any([i is None for i in [x, y, z]]):
            return None

        phi = np.arctan2(y, x)
        rho = np.sqrt(x ** 2 + y ** 2)
        r = np.sqrt(x ** 2 + y ** 2 + z ** 2)
        if r != 0:
            theta = np.arccos(z / r)
        else:
            theta = 0

        return x, y, z, r, theta, phi, rho

    @staticmethod
    def _spherical_to_other(r, theta, phi):
        """Convert from spherical to other representations."""
        if any([i is None for i in [r, theta, phi]]):
            return None

        z = r * np.cos(theta)
        x = r * np.sin(theta) * np.cos(phi)
        y = r * np.sin(theta) * np.sin(phi)
        rho = np.sqrt(x ** 2 + y ** 2)

        return x, y, z, r, theta, phi, rho

    @staticmethod
    def _cylindrical_to_other(phi, rho, z):
        """Convert from cylindrical to other representations."""
        if any([i is None for i in [phi, rho, z]]):
            return None

        x = rho * np.cos(phi)
        y = rho * np.sin(phi)
        r = np.sqrt(rho ** 2 + z ** 2)
        if r != 0:
            theta = np.arccos(z / r)
        else:
            theta = 0

        return x, y, z, r, theta, phi, rho

    def _compute_unknowns(self) -> None:
        """
        Compute all coordinates. To do this we need either the set (x, y, z)
        to contain no ``None`` values, or the set (r, theta, phi), or the set
        (rho, phi, z). Given any of these sets, we can recompute the rest.

        This function will raise an error if there are contradictory inputs
        (e.g. x=3, y=4, z=0 and rho=6).
        """
        for f in [
            lambda: FieldVector._cartesian_to_other(self._x, self._y, self._z),
            lambda: FieldVector._spherical_to_other(self._r, self._theta,
                                                    self._phi),
            lambda: FieldVector._cylindrical_to_other(self._phi, self._rho,
                                                      self._z)
        ]:
            new_values = f()
            if new_values is not None:  # this will return None if any of the
                # function arguments is None.
                self._set_attribute_values(FieldVector.attributes, new_values)
                break

    def copy(self: T, other: T) -> None:
        """Copy the properties of other vector to yourself."""
        for att in FieldVector.attributes:
            value = getattr(other, "_" + att)
            setattr(self, "_" + att, value)

    def set_vector(self, **new_values: float) -> None:
        """
        Reset the the values of the vector.

        Examples:
            >>> f = FieldVector(x=0, y=2, z=6)
            >>> f.set_vector(x=9, y=3, z=1)
            >>> f.set_vector(r=1, theta=30.0, phi=10.0)
            # The following should raise a value error:
            # "Can only set vector with a complete value set"
            >>> f.set_vector(x=9, y=0)
            # Although mathematically it is possible to compute the complete
            # vector from the values given, this is too hard to implement with
            # generality (and not worth it), so the following will raise the
            # above-mentioned ValueError too.
            >>> f.set_vector(x=9, y=0, r=3)
        """
        names = sorted(list(new_values.keys()))
        groups = [["x", "y", "z"], ["phi", "r", "theta"], ["phi", "rho", "z"]]
        if names not in groups:
            raise ValueError("Can only set vector with a complete value set")

        new_vector = FieldVector(**new_values)
        self.copy(new_vector)

    def set_component(self, **new_values: float) -> None:
        """
        Set a single component of the vector to some new value. It is
        disallowed for the user to set vector components manually as this can
        lead to inconsistencies (e.g. x and rho are not independent of each
        other, setting one has to effect the other).

        Examples:
            >>> f = FieldVector(x=2, y=3, z=4)
            # Since r is part of the set (r, theta, phi) representing
            # spherical coordinates, setting r means that theta and phi are
            # kept constant and only r is changed. After changing r,
            # (x, y, z) values are recomputed, as is the rho coordinate.
            # Internally we arrange this by setting x, y, z and rho to None
            # and calling self._compute_unknowns().
            >>> f.set_component(r=10)

        Args:
            new_values (dict): Keys representing parameter names and values the
                values to be set.
        """
        if len(new_values) > 1:
            raise NotImplementedError("Cannot set multiple components at once")

        items = list(new_values.items())
        component_name = items[0][0]

        if component_name in ["theta", "phi"]:
            # convert angles to radians
            value = np.radians(items[0][1])
        else:
            value = items[0][1]

        setattr(self, "_" + component_name, float(value))

        groups = [["x", "y", "z"], ["r", "theta", "phi"], ["phi", "rho", "z"]]

        for group in groups:
            if component_name in group:

                for att in FieldVector.attributes:
                    if att not in group:
                        setattr(self, "_" + att, None)

                break

        self._compute_unknowns()

    def get_components(self, *names: str):
        """Get field components by name."""

        def convert_angle_to_degrees(name: str, value: float) -> float:
            # Convert all angles to degrees
            if name in ["theta", "phi"]:
                return float(np.degrees(value))
            else:
                return value

        components = [convert_angle_to_degrees(
            name, getattr(self, "_" + name)
        ) for name in names]

        return components

    def is_equal(self, other: "FieldVector") -> bool:
        """
        Returns ``True`` if ``other`` is equivalent to ``self``, ``False`` otherwise.
        """
        for name in ["x", "y", "z"]:
            self_value = getattr(self, name)
            other_value = getattr(other, name)
            if not np.isclose(self_value, other_value):
                return False

        return True

    def __getitem__(self, component: str) -> float:
        return self.get_components(component)[0]

    def __setitem__(self, component: str, value: float) -> None:
        self.set_component(**{component: value})

    def __mul__(self, other: Any) -> "FieldVector":
        if not isinstance(other, (float, int)):
            return NotImplemented

        return FieldVector(**{
            component: self[component] * other
            for component in 'xyz'
        })

    def __rmul__(self, other: Any) -> "FieldVector":
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self * other

    def __truediv__(self, other: Any) -> "FieldVector":
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self * (1.0 / other)

    def __neg__(self) -> "FieldVector":
        return -1 * self

    def __add__(self, other: Any) -> "FieldVector":
        if not isinstance(other, FieldVector):
            return NotImplemented

        return FieldVector(**{
            component: self[component] + other[component]
            for component in 'xyz'
        })

    def __sub__(self, other: Any) -> "FieldVector":
        if not isinstance(other, FieldVector):
            return NotImplemented

        return FieldVector(**{
            component: self[component] - other[component]
            for component in 'xyz'
        })

    # NB: we disable the pylint warning here so that we can match
    #     NumPy's naming convention for the norm method.
    def norm(self,
             ord: NormOrder = 2  # pylint: disable=redefined-builtin
             ) -> float:
        """
        Returns the norm of this field vector. See ``np.norm``
        for the definition of the ord keyword argument.
        """
        return np.linalg.norm([self.x, self.y, self.z], ord=ord)

    def distance(self, other,
                 ord: NormOrder = 2  # pylint: disable=redefined-builtin
                 ) -> float:
        return (self - other).norm(ord=ord)

    @property
    def x(self) -> Optional[float]:
        return self._x

    @property
    def y(self) -> Optional[float]:
        return self._y

    @property
    def z(self) -> Optional[float]:
        return self._z

    @property
    def rho(self) -> Optional[float]:
        return self._rho

    @property
    def theta(self) -> Optional[float]:
        if self._theta is None:
            return None
        return float(np.degrees(self._theta))

    @property
    def r(self) -> Optional[float]:
        return self._r

    @property
    def phi(self) -> Optional[float]:
        if self._phi is None:
            return None
        return float(np.degrees(self._phi))

    # Representation Methods #

    def repr_cartesian(self) -> str:
        return f"FieldVector(x={self.x}, y={self.y}, z={self.z})"

    def repr_spherical(self) -> str:
        return f"FieldVector(r={self.r}, phi={self.phi}, theta={self.theta})"

    def repr_cylindrical(self) -> str:
        return f"FieldVector(rho={self.rho}, phi={self.phi}, z={self.z})"

    def __repr__(self) -> str:
        if self.repr_format == "cartesian":
            return self.repr_cartesian()
        elif self.repr_format == "spherical":
            return self.repr_spherical()
        elif self.repr_format == "cylindrical":
            return self.repr_cylindrical()
        else:
            return super().__repr__()

    # Homogeneous Coordinates #

    def as_homogeneous(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z, 1])

    @classmethod
    def from_homogeneous(cls: Type[T], hvec: np.ndarray) -> T:
        # Homogeneous coordinates define an equivalence relation
        #     [x / s, y / s, z / s, 1] == [x, y, z, s].
        # More generally,
        #     [x, y, z, s] == [x', y', z', s']
        #     iff x / s == x' / s',
        #         y / s == y' / s', and
        #         z / s == z' / s'.
        # This definition has the consequence that for any w,
        # w * [x, y, z, s] == [x, y, z, s].
        # Thus, we start by rescaling such that s == 1.
        hvec /= hvec[-1]
        return cls(
            x=hvec[0], y=hvec[1], z=hvec[2]
        )



from typing import Any

import pytest
from qcodes.utils.validators import Anything, Validator

from .conftest import AClass, a_func


class BrokenValidator(Validator[Any]):

    def __init__(self):
        pass


def test_broken():
    # nor can you call validate without overriding it in a subclass
    b = BrokenValidator()
    with pytest.raises(NotImplementedError):
        b.validate(0)


def test_real_anything():
    a = Anything()
    for v in [None, 0, 1, 0.0, 1.2, '', 'hi!', [1, 2, 3], [],
              {'a': 1, 'b': 2}, {}, {1, 2, 3}, a, range(10),
              True, False, float("nan"), float("inf"), b'good',
              AClass, AClass(), a_func]:
        a.validate(v)

    assert repr(a) == '<Anything>'


def test_failed_anything():
    with pytest.raises(TypeError):
        Anything(1)

    with pytest.raises(TypeError):
        Anything(values=[1, 2, 3])



import numpy as np
import pytest
from qcodes.utils.validators import PermissiveMultiples

divisors = [40e-9, -1, 0.2225, 1 / 3, np.pi / 2]

multiples = [[800e-9, -40e-9, 0, 1],
             [3, -4, 0, -1, 1],
             [1.5575, -167.9875, 0],
             [2 / 3, 3, 1, 0, -5 / 3, 1e4],
             [np.pi, 5 * np.pi / 2, 0, -np.pi / 2]]

not_multiples = [[801e-9, 4.002e-5],
                 [1.5, 0.9999999],
                 [0.2226],
                 [0.6667, 28 / 9],
                 [3 * np.pi / 4]]


def test_passing():
    for divind, div in enumerate(divisors):
        val = PermissiveMultiples(div)
        for mult in multiples[divind]:
            val.validate(mult)


def test_not_passing():
    for divind, div in enumerate(divisors):
        val = PermissiveMultiples(div)
        for mult in not_multiples[divind]:
            with pytest.raises(ValueError):
                val.validate(mult)


# finally, a quick test that the precision is indeed setable
def test_precision():
    pm_lax = PermissiveMultiples(35e-9, precision=3e-9)
    pm_lax.validate(72e-9)
    pm_strict = PermissiveMultiples(35e-9, precision=1e-10)
    with pytest.raises(ValueError):
        pm_strict.validate(70.2e-9)


def test_valid_values():
    for div in divisors:
        pm = PermissiveMultiples(div)
        for val in pm.valid_values:
            pm.validate(val)



import logging
import re
import time
from functools import partial

import numpy as np

from qcodes import IPInstrument, MultiParameter
from qcodes.utils.deprecate import deprecate
from qcodes.utils.validators import Bool, Enum

log = logging.getLogger(__name__)


@deprecate(alternative="MercuryiPS_VISA.MercuryiPS")
class MercuryiPSArray(MultiParameter):
    """
    This parameter holds the MercuryiPS's 3 dimensional parameters

    """
    def __init__(self, name, instrument, names, units, get_cmd, set_cmd, **kwargs):
        shapes = tuple(() for i in names)
        super().__init__(
            name, names, shapes, snapshot_value=True, instrument=instrument, **kwargs
        )
        self._get = get_cmd
        self._set = set_cmd
        self.units = units

    def get_raw(self):
        try:
            value = self._get()
            return value
        except Exception as e:
            e.args = e.args + (f'getting {self.full_name}',)
            raise e

    def set_raw(self, setpoint):
        return self._set(setpoint)


@deprecate(alternative="MercuryiPS_VISA.MercuryiPS")
class MercuryiPS(IPInstrument):
    """
    This is the qcodes driver for the Oxford MercuryiPS magnet power supply.

    Args:
        name (str): name of the instrument
        address (str): The IP address or domain name of this instrument
        port (int): the IP port to communicate on (default 7020, as in the manual)

        axes (List[str], Optional): axes to support, as a list of uppercase
            characters, eg ``['X', 'Y', 'Z']``. If omitted, will ask the
            instrument what axes it supports.

    Status: beta-version.

    .. todo::

        - SAFETY!! we need to make sure the magnet is only ramped at certain
          conditions!
        - make ATOB a parameter, and move all possible to use
          _read_cmd, _write_cmd
        - this findall stuff in _get_cmd, is that smart?

    The driver is written as an IPInstrument, but it can likely be converted to
    ``VisaInstrument`` by removing the ``port`` arg and defining methods:

        - ``def _send(self, msg): self.visa_handle.write(msg)``
        - ``def _recv(self): return self.visa_handle.read()``

    """
    # def __init__(self, name, axes=None, **kwargs):
    #     super().__init__(name, terminator='\n', **kwargs)
    def __init__(self, name, address=None, port=7020, axes=None, **kwargs):
        super().__init__(name, address=address, port=port, terminator='\n',
                         **kwargs)
        self.axes = axes
        self._ATOB = []
        self._latest_response = ''
        # for some reason the first call is always invalid?!
        # need some kind of init?
        self.ask('*IDN?')

        if axes is None:
            self._determine_magnet_axes()
        self._determine_current_to_field()


        self.add_parameter('hold_after_set',
                           get_cmd=None, set_cmd=None,
                           vals=Bool(),
                           initial_value=True,
                           docstring='Should the driver block while waiting for the Magnet power supply '
                                     'to go into hold mode.'
                           )

        self.add_parameter('setpoint',
                           names=tuple('B' + ax.lower() + '_setpoint' for ax in self.axes),
                           units=tuple('T' for ax in self.axes),
                           get_cmd=partial(self._get_fld, self.axes, 'FSET'),
                           set_cmd=partial(self._ramp_to_setpoint, self.axes, 'FSET'),
                           parameter_class=MercuryiPSArray)

        self.add_parameter('rate',
                           names=tuple('rate_B' + ax.lower() for ax in self.axes),
                           units=tuple('T/m' for ax in self.axes),
                           get_cmd=partial(self._get_fld, self.axes, 'RFST'),
                           set_cmd=partial(self._ramp_to_setpoint, self.axes, 'RFST'),
                           parameter_class=MercuryiPSArray)

        self.add_parameter('fld',
                           names=tuple('B'+ax.lower() for ax in self.axes),
                           units=tuple('T'for ax in self.axes),
                           get_cmd=partial(self._get_fld, self.axes, 'FLD'),
                           set_cmd=partial(self._ramp_to_setpoint, self.axes, 'FSET'),
                           parameter_class=MercuryiPSArray)

        self.add_parameter('fldC',
                           names=['B'+ax.lower() for ax in self.axes],
                           units=['T' for ax in self.axes],
                           get_cmd=partial(self._get_fld, self.axes, 'CURR'),
                           set_cmd=partial(self._ramp_to_setpoint, self.axes, 'CSET'),
                           parameter_class=MercuryiPSArray)

        self.add_parameter('rtp',
                           names=['radius', 'theta', 'phi'],
                           units=['|B|', 'rad', 'rad'],
                           get_cmd=partial(self._get_rtp, self.axes, 'FLD'),
                           set_cmd=partial(self._set_rtp, self.axes, 'FSET'),
                           parameter_class=MercuryiPSArray)

        self.add_parameter('rtpC',
                           names=['radius', 'theta', 'phi'],
                           units=['|B|', 'rad', 'rad'],
                           get_cmd=partial(self._get_rtp, self.axes, 'CURR'),
                           set_cmd=partial(self._set_rtp, self.axes, 'CSET'),
                           parameter_class=MercuryiPSArray)

        # so we have radius, theta and phi in buffer
        self.rtp.get()

        self.add_parameter('radius',
                           get_cmd=self._get_r,
                           set_cmd=self._set_r,
                           unit='|B|')
        self.add_parameter('theta',
                           get_cmd=self._get_theta,
                           set_cmd=self._set_theta,
                           unit='rad')
        self.add_parameter('phi',
                           get_cmd=self._get_phi,
                           set_cmd=self._set_phi,
                           unit='rad')

        for ax in self.axes:
            self.add_parameter(ax.lower()+'_fld',
                               get_cmd=partial(self._get_fld, ax, 'FLD'),
                               set_cmd=partial(self._ramp_to_setpoint,
                                               ax, 'FSET'),
                               label='B'+ax.lower(),
                               unit='T')
            self.add_parameter(ax.lower()+'_fldC',
                               get_cmd=partial(self._get_fld,
                                               ax, 'CURR'),
                               set_cmd=partial(self._ramp_to_setpoint,
                                               ax, 'CSET'),
                               label='B'+ax.lower(),
                               unit='T')
            self.add_parameter(ax.lower()+'_fld_wait',
                               get_cmd=partial(self._get_fld,
                                               ax, 'CURR'),
                               set_cmd=partial(self._ramp_to_setpoint_and_wait,
                                               ax, 'CSET'),
                               label='B'+ax.lower(),
                               unit='T')
            self.add_parameter(ax.lower()+'_ACTN',
                               get_cmd=partial(
                                   self._get_cmd,
                                   'READ:DEV:GRP' + ax + ':PSU:ACTN?'),
                               set_cmd='SET:DEV:GRP' + ax + ':PSU:ACTN:{}',
                               vals=Enum('HOLD', 'RTOS', 'RTOZ', 'CLMP'))
            self.add_parameter(ax.lower()+'_setpoint',
                               get_cmd=partial(self._get_fld, ax, 'FSET'),
                               set_cmd=partial(self._set_fld, ax, 'FSET'),
                               unit='T')
            self.add_parameter(ax.lower()+'_setpointC',
                               get_cmd=partial(self._get_fld, ax, 'CSET'),
                               set_cmd=partial(self._set_fld, ax, 'CSET'),
                               unit='T')
            self.add_parameter(ax.lower()+'_rate',
                               get_cmd=partial(self._get_fld, ax, 'RFST'),
                               set_cmd=partial(self._set_fld, ax, 'RFST'),
                               unit='T/m')
            self.add_parameter(ax.lower()+'_rateC',
                               get_cmd=partial(self._get_fld, ax, 'RCST'),
                               set_cmd=partial(self._set_fld, ax, 'RCST'),
                               unit='T/m')

            self.connect_message()

    def hold(self):
        for ax in self.axes:
            actn = self.parameters[ax.lower() + '_ACTN'].get()
            if actn == 'CLMP':
                log.info("Skipping: Trying to set clamped axis {} to HOLD: "
                         "If you really want to HOLD the clamped axis use "
                         "mercury.ACTN('HOLD')".format(ax))
            else:
                self.parameters[ax.lower() + '_ACTN'].set('HOLD')

    def rtos(self):
        for ax in self.axes:
            actn = self.parameters[ax.lower() + '_ACTN'].get()
            if actn == 'CLMP':
                log.info("Skipping: Trying to set clamped axis {} to RTOS: "
                            "If you really want to use the clamped axis use "
                            "mercury.{}_ACTN('HOLD') to unclamp"
                            "first".format(ax, ax))
            else:
                self.parameters[ax.lower() + '_ACTN'].set('RTOS')

    def to_zero(self):
        for ax in self.axes:
            actn = self.parameters[ax.lower() + '_ACTN'].get()
            if actn == 'CLMP':
                log.info("Skipping: Trying to set clamped axis {} to zero:"
                         "If you really want to use the clamped axis use "
                         "mercury.{}_ACTN('HOLD') to unclamp"
                         "first".format(ax, ax))
            else:
                self.parameters[ax.lower() + '_ACTN'].set('RTOZ')

    def _ramp_to_setpoint(self, ax, cmd, setpoint):
        if 'CLMP' in [getattr(self, a.lower() + '_ACTN')() for a in ax]:
            raise RuntimeError('Trying to ramp a clamped axis. Please unclamp first')
        self._set_fld(ax, cmd, setpoint)
        for a in ax:
            getattr(self, a.lower() + '_ACTN')('RTOS')

        if self.hold_after_set():
            try:
                while not all([getattr(self, a.lower() + '_ACTN')() in ['HOLD', 'CLMP'] for a in ax]):
                    time.sleep(0.1)
            except KeyboardInterrupt:
                self.hold()
                raise KeyboardInterrupt
            finally:
                if 'CLMP' in [getattr(self, a.lower() + '_ACTN')() for a in ax]:
                    log.warning("One or more of the ramped axis have been clamped")

    def _ramp_to_setpoint_and_wait(self, ax, cmd, setpoint):
        error = 0.2e-3
        fldc = self.parameters[ax.lower()+'_fldC']
        fldc.set(setpoint)

        while abs(fldc.get() - setpoint) > error:
            time.sleep(0.5)
        if setpoint == 0.0:
            # This ensures that the magnet wont try to go hold
            # some funny xe-5 value.
            fld = getattr(self, ax.lower()+'_fld')
            fld.set(setpoint)

    def _set_fld(self, ax, cmd, setpoint):
        # Could be FSET for setpoint
        #          RFST for rate
        if not isinstance(setpoint, list):
            setpoint = [setpoint]
        if cmd in ['CSET', 'RCST', 'CURR', 'PCUR', 'RCUR']:
            setpoint = np.array(self._ATOB) * np.array(setpoint)
            if len(ax) == 1:
                setpoint = setpoint[self.axes.index(ax)]
        elif len(ax) == 1:
            setpoint = setpoint[0]

        msg = 'SET:DEV:GRP{}:PSU:SIG:{}:{:6f}'
        self._write_cmd(cmd, ax, setpoint, msg)

    def _get_fld(self, ax, cmd):
        # Could be FSET for setpoint
        #          FLD for field
        #          RFLD for rate
        #          PFLD persistent field reading
        fld = list(self._read_cmd(cmd, ax, float))

        if cmd in ['CSET', 'RCST', 'CURR', 'PCUR', 'RCUR']:
            fld = np.array(fld) / np.array(self._ATOB)
            if len(ax) == 1:
                return fld[self.axes.index(ax)]
        elif len(ax) == 1:
            return fld[0]
        return list(fld)

    def _get_rtp(self, ax, cmd):
        fld = self._get_fld(ax, cmd)
        sphere = self._carttosphere(fld)
        self._radius, self._theta, self._phi = sphere
        return sphere

    def _set_rtp(self, ax, cmd, setpoint):
        fld = self._spheretocart(setpoint)
        self._ramp_to_setpoint(ax, cmd, fld)

    def _get_r(self):
        self.rtp.get()
        return self._radius

    def _set_r(self, val):
        self.rtp.set([val, self._theta, self._phi])

    def _get_theta(self):
        self.rtp.get()
        return self._theta

    def _set_theta(self, val):
        self.rtp.set([self._radius, val, self._phi])

    def _get_phi(self):
        self.rtp.get()
        return self._phi

    def _set_phi(self, val):
        self.rtpC.set([self._radius, self._theta, val])

    def _determine_magnet_axes(self):
        cat = self.ask('READ:SYS:CAT')
        self.axes = re.findall('DEV:GRP(.+?):PSU', cat)

    def _determine_current_to_field(self):
        # This has a unit A/T
        self._ATOB = []
        for ax in self.axes:
            r = self._get_cmd(f'READ:DEV:GRP{ax}:PSU:ATOB?', float)
            self._ATOB.append(r)

    def _read_cmd(self, cmd, axes, parser=None, fmt=None):
        fmt = fmt or 'READ:DEV:GRP{}:PSU:SIG:{}?'
        msg = ''
        msglist = []
        for axis in axes:
            msglist.append(fmt.format(axis, cmd))
        msg = '\n'.join(msglist)
        log.info(f"Writing '{msg}' to Mercury")
        self._send(msg)
        rep = self._recv()
        log.info(f"Read '{rep}' from Mercury")
        data = [None] * len(axes)
        for i in range(20):
            for ln in rep.split('\n'):
                for ix, msg in enumerate(msglist):
                    if msg[5:-1] in ln:
                        val = ln.split(':')[-1]
                        if parser is float:
                            try:
                                matches = re.findall(r"[-+]?\d*\.\d+|\d+", val)
                                val = float(matches[0])
                            except:
                                continue
                        data[ix] = val
                    if not (None in data):
                        return data
            rep = self._recv()
            log.info(f"Read '{rep}' from Mercury")
        return data

    def _write_cmd(self, cmd, axes, setpoint, fmt=None, parser=None):
        fmt = fmt or 'SET:DEV:GRP{}:PSU:SIG:{}:{:4f}'
        msg = ''
        msglist = []
        if len(axes) == 1:
            setpoint = [setpoint]
        for ix, axis in enumerate(axes):
            msglist.append(fmt.format(axis, cmd, setpoint[ix]))
        msg = '\n'.join(msglist)
        log.info(f"Writing '{msg}' to Mercury")
        self._send(msg)
        rep = self._recv()
        log.info(f"Read '{rep}' from Mercury")
        data = [None] * len(axes)
        for i in range(20):
            for ln in rep.split('\n'):
                for ix, msg in enumerate(msglist):
                    if msg[5:-1] in ln:
                        val = ln.split(':')[-1]
                        if parser is float:
                            try:
                                matches = re.findall(r"[-+]?\d*\.\d+|\d+", val)
                                val = float(matches[0])
                            except:
                                continue
                        data[ix] = val
                    if not (None in data):
                        return data
            rep = self._recv()
            log.info(f"Read '{rep}' from Mercury")

    def _get_cmd(self, question, parser=None):
        log.info(f"Writing '{question}' to Mercury")
        rep = self.ask(question)
        log.info(f"Read '{rep}' from Mercury")
        self._latest_response = rep
        msg = rep[len(question):]
        # How would one match this without specifying the units?
        # m = re.match('STAT:DEV:GRPX:PSU:SIG:RFST:(.+?)T/m',
        #              'STAT:DEV:GRPX:PSU:SIG:RFST:0.0200T/m')
        # m.groups()[0]
        if parser is float:
            try:
                return(float(re.findall(r"[-+]?\d*\.\d+|\d+", msg)[0]))
            except:
                return None
        return msg.strip()

    def write(self, msg):
        log.info(f"Writing '{msg}' to Mercury")
        rep = self.ask(msg)
        log.info(f"Read '{rep}' from Mercury")
        self._latest_response = rep
        if 'INVALID' in rep:
            print('warning', msg, rep)

    def _spheretocart(self, sphere):
        """
        r,  theta,  phi = sphere
        """
        r,  theta,  phi = sphere
        x = (r * np.sin(theta) * np.cos(phi))
        y = (r * np.sin(theta) * np.sin(phi))
        z = (r * np.cos(theta))
        return [x,  y,  z]

    def _carttosphere(self, field):
        field = np.array(field)
        r = np.sqrt(np.sum(field**2))
        if r == 0:
            theta = 0
            phi = 0
        else:
            theta = np.arccos(field[2] / r)
            phi = np.arctan2(field[1],  field[0])
            if phi<0:
                phi = phi+np.pi*2
        return [r, theta, phi]



import logging
from functools import partial
from typing import Any, Callable, Dict, Optional, Sequence

import numpy as np

from qcodes import (
    ArrayParameter,
    ChannelList,
    Instrument,
    InstrumentChannel,
    VisaInstrument,
)
from qcodes import validators as vals
from qcodes.instrument.parameter import ParamRawDataType
from qcodes.utils.validators import Enum, Numbers

log = logging.getLogger(__name__)


class TraceNotReady(Exception):
    pass


class TraceSetPointsChanged(Exception):
    pass


class RawTrace(ArrayParameter):
    """
    raw_trace will return a trace from OSCIL
    """

    def __init__(self, name: str, instrument: Instrument, channel: int):
        super().__init__(name,
                         shape=(1024,),
                         label='Voltage',
                         unit='V',
                         setpoint_names=('Time',),
                         setpoint_labels=(
                             f'Channel {channel} time series',),
                         setpoint_units=('s',),
                         docstring='raw trace from the scope',
                         instrument=instrument
                         )
        self._channel = channel

    def prepare_curvedata(self) -> None:
        """
        Prepare the scope for returning curve data
        """
        # To calculate set points, we must have the full preamble
        # For the instrument to return the full preamble, the channel
        # in question must be displayed

        # shorthand
        instr = self.instrument
        assert isinstance(instr, InfiniiumChannel)

        # number of set points
        self.npts = int(instr.ask("WAV:POIN?"))
        # first set point
        self.xorigin = float(instr.ask(":WAVeform:XORigin?"))
        # step size
        self.xincrem = float(instr.ask(":WAVeform:XINCrement?"))
        # calculate set points
        xdata = np.linspace(self.xorigin,
                            self.npts * self.xincrem + self.xorigin, self.npts)

        # set setpoints
        self.setpoints = (tuple(xdata), )
        self.shape = (self.npts, )

        # set up the instrument
        # ---------------------------------------------------------------------

        # make this on a per channel basis?
        root_instrument = instr.root_instrument
        assert isinstance(root_instrument, Infiniium)
        root_instrument.trace_ready = True

    def get_raw(self) -> ParamRawDataType:
        # when get is called the setpoints have to be known already
        # (saving data issue). Therefor create additional prepare function that
        # queries for the size.
        # check if already prepared
        instr = self.instrument
        assert isinstance(instr, InfiniiumChannel)

        if not instr.root_instrument.trace_ready:
            raise TraceNotReady('Please run prepare_curvedata to prepare '
                                'the scope for acquiring a trace.')

        # shorthand

        # set up the instrument
        # ---------------------------------------------------------------------

        # TODO: check number of points
        # check if requested number of points is less than 500 million

        # get intrument state
        state = instr.ask(':RSTate?')

        # acquire the data
        # ---------------------------------------------------------------------

        # digitize is the actual call for acquisition, blocks
        instr.write(f':DIGitize CHANnel{self._channel}')

        # transfer the data
        # ---------------------------------------------------------------------

        # select the channel from which to read
        instr._parent.data_source(f'CHAN{self._channel}')
        # specifiy the data format in which to read
        instr.write(':WAVeform:FORMat WORD')
        instr.write(":waveform:byteorder LSBFirst")
        # streaming is only required for data > 1GB
        instr.write(':WAVeform:STReaming OFF')

        # request the actual transfer
        data = instr._parent.visa_handle.query_binary_values(
            'WAV:DATA?', datatype='h', is_big_endian=False,
            expect_termination=False)
        # the Infiniium does not include an extra termination char on binary
        # messages so we set expect_termination to False

        if len(data) != self.shape[0]:
            raise TraceSetPointsChanged('{} points have been aquired and {} \
            set points have been prepared in \
            prepare_curvedata'.format(len(data), self.shape[0]))
        # check x data scaling
        xorigin = float(instr.ask(":WAVeform:XORigin?"))
        # step size
        xincrem = float(instr.ask(":WAVeform:XINCrement?"))
        error = self.xorigin - xorigin
        # this is a bad workaround
        if error > xincrem:
            raise TraceSetPointsChanged('{} is the prepared x origin and {} \
            is the x origin after the measurement.'.format(self.xorigin,
                                                           xorigin))
        error = (self.xincrem - xincrem) / xincrem
        if error > 1e-6:
            raise TraceSetPointsChanged('{} is the prepared x increment and {} \
            is the x increment after the measurement.'.format(self.xincrem,
                                                              xincrem))
        # y data scaling
        yorigin = float(instr.ask(":WAVeform:YORigin?"))
        yinc = float(instr.ask(":WAVeform:YINCrement?"))
        channel_data = np.array(data)
        channel_data = np.multiply(channel_data, yinc) + yorigin

        # restore original state
        # ---------------------------------------------------------------------

        # switch display back on
        instr.write(f':CHANnel{self._channel}:DISPlay ON')
        # continue refresh
        if state == 'RUN':
            instr.write(':RUN')

        return channel_data


class MeasurementSubsystem(InstrumentChannel):
    """
    Submodule containing the measurement subsystem commands and associated
    parameters
    """
    # note: this is not really a channel, but InstrumentChannel does everything
    # a 'Submodule' class should do

    def __init__(self, parent: Instrument, name: str, **kwargs: Any) -> None:
        super().__init__(parent, name, **kwargs)

        self.add_parameter(name='source_1',
                           label='Measurement primary source',
                           set_cmd=partial(self._set_source, 1),
                           get_cmd=partial(self._get_source, 1),
                           val_mapping={i: f'CHAN{i}' for i in range(1, 5)},
                           snapshot_value=False)

        self.add_parameter(name='source_2',
                           label='Measurement secondary source',
                           set_cmd=partial(self._set_source, 2),
                           get_cmd=partial(self._get_source, 2),
                           val_mapping={i: f'CHAN{i}' for i in range(1, 5)},
                           snapshot_value=False)

        self.add_parameter(name='amplitude',
                           label='Voltage amplitude',
                           get_cmd=self._make_meas_cmd('VAMPlitude'),
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='average',
                           label='Voltage average',
                           get_cmd=self._make_meas_cmd('VAVerage'),
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='base',
                           label='Statistical base',
                           get_cmd=self._make_meas_cmd('VBASe'),
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='frequency',
                           label='Signal frequency',
                           get_cmd=self._make_meas_cmd('FREQuency'),
                           get_parser=float,
                           unit='Hz',
                           docstring="""
                                     measure the frequency of the first
                                     complete cycle on the screen using
                                     the mid-threshold levels of the waveform
                                     """,
                           snapshot_value=False)

        self.add_parameter(name='lower',
                           label='Voltage lower',
                           get_cmd=self._make_meas_cmd('VLOWer'),
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='max',
                           label='Voltage maximum',
                           get_cmd=self._make_meas_cmd('VMAX'),
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='middle',
                           label='Middle threshold voltage',
                           get_cmd=self._make_meas_cmd('VMIDdle'),
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='min',
                           label='Voltage minimum',
                           get_cmd=self._make_meas_cmd('VMIN'),
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='overshoot',
                           label='Voltage overshoot',
                           get_cmd=self._make_meas_cmd('VOVershoot'),
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='vpp',
                           label='Voltage peak-to-peak',
                           get_cmd=self._make_meas_cmd('VPP'),
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='rms',
                           label='Voltage RMS',
                           get_cmd=self._make_meas_cmd('VRMS') + ' DISPlay, DC',
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

        self.add_parameter(name='rms_no_DC',
                           label='Voltage RMS',
                           get_cmd=self._make_meas_cmd('VRMS') + ' DISPlay, AC',
                           get_parser=float,
                           unit='V',
                           snapshot_value=False)

    @staticmethod
    def _make_meas_cmd(cmd: str) -> str:
        """
        Helper function to avoid typos
        """
        return f':MEASure:{cmd}?'

    def _set_source(self, rank: int, source: str) -> None:
        """
        Set the measurement source, either primary (rank==1) or secondary
        (rank==2)
        """
        sources = self.ask(':MEASure:SOURCE?').split(',')
        if rank == 1:
            self.write(f':MEASure:SOURCE {source}, {sources[1]}')
        else:
            self.write(f':MEASure:SOURCE {sources[0]}, {source}')

    def _get_source(self, rank: int) -> str:
        """
        Get the measurement source, either primary (rank==1) or secondary
        (rank==2)
        """
        sources = self.ask(':MEASure:SOURCE?').split(',')

        return sources[rank-1]


class InfiniiumChannel(InstrumentChannel):

    def __init__(self, parent: "Infiniium", name: str, channel: int):
        super().__init__(parent, name)
        # display
        self.add_parameter(name='display',
                           label=f'Channel {channel} display on/off',
                           set_cmd=f'CHANnel{channel}:DISPlay {{}}',
                           get_cmd=f'CHANnel{channel}:DISPlay?',
                           val_mapping={True: 1, False: 0},
                           )
        # scaling
        self.add_parameter(name='offset',
                           label=f'Channel {channel} offset',
                           set_cmd=f'CHAN{channel}:OFFS {{}}',
                           unit='V',
                           get_cmd=f'CHAN{channel}:OFFS?',
                           get_parser=float
                           )

        # scale and range are interdependent, when setting one, invalidate the
        # the other.
        # Scale is commented out for this reason
        # self.add_parameter(name='scale',
        #                    label='Channel {} scale'.format(channel),
        #                    unit='V/div',
        #                    set_cmd='CHAN{}:SCAL {{}}'.format(channel),
        #                    get_cmd='CHAN{}:SCAL?'.format(channel),
        #                    get_parser=float,
        #                    vals=vals.Numbers(0,100)  # TODO: upper limit?
        #                    )

        self.add_parameter(name='range',
                           label=f'Channel {channel} range',
                           unit='V',
                           set_cmd=f'CHAN{channel}:RANG {{}}',
                           get_cmd=f'CHAN{channel}:RANG?',
                           get_parser=float,
                           vals=vals.Numbers()
                           )
        # trigger
        self.add_parameter(
            'trigger_level',
            label=f'Tirgger level channel {channel}',
            unit='V',
            get_cmd=f':TRIGger:LEVel? CHANnel{channel}',
            set_cmd=f':TRIGger:LEVel CHANnel{channel},{{}}',
            get_parser=float,
            vals=Numbers(),
        )

        # Acquisition
        self.add_parameter(name='trace',
                           channel=channel,
                           parameter_class=RawTrace
                           )


class Infiniium(VisaInstrument):
    """
    This is the QCoDeS driver for the Keysight Infiniium oscilloscopes from the
     - tested for MSOS104A of the Infiniium S-series.
    """

    def __init__(self, name: str, address: str,
                 timeout: float = 20,
                 **kwargs: Any):
        """
        Initialises the oscilloscope.

        Args:
            name: Name of the instrument used by QCoDeS
            address: Instrument address as used by VISA
            timeout: visa timeout, in secs.
        """

        super().__init__(name, address, timeout=timeout,
                         terminator='\n', **kwargs)
        self.connect_message()

        # Scope trace boolean
        self.trace_ready = False

        # switch the response header off,
        # else none of our parameters will work
        self.write(':SYSTem:HEADer OFF')

        # functions

        # general parameters

        # the parameters are in the same order as the front panel.
        # Beware, he list of implemented parameters is not complete. Refer to
        # the manual (Infiniium prog guide) for an equally infiniium list.

        # time base

        # timebase_scale is commented out for same reason as channel scale
        # use range instead
        # self.add_parameter('timebase_scale',
        #                    label='Scale of the one time devision',
        #                    unit='s/Div',
        #                    get_cmd=':TIMebase:SCALe?',
        #                    set_cmd=':TIMebase:SCALe {}',
        #                    vals=Numbers(),
        #                    get_parser=float,
        #                    )

        self.add_parameter('timebase_range',
                           label='Range of the time axis',
                           unit='s',
                           get_cmd=':TIMebase:RANGe?',
                           set_cmd=':TIMebase:RANGe {}',
                           vals=Numbers(5e-12, 20),
                           get_parser=float,
                           )
        self.add_parameter('timebase_position',
                           label='Offset of the time axis',
                           unit='s',
                           get_cmd=':TIMebase:POSition?',
                           set_cmd=':TIMebase:POSition {}',
                           vals=Numbers(),
                           get_parser=float,
                           )

        self.add_parameter('timebase_roll_enabled',
                           label='Is rolling mode enabled',
                           get_cmd=':TIMebase:ROLL:ENABLE?',
                           set_cmd=':TIMebase:ROLL:ENABLE {}',
                           val_mapping={True: 1, False: 0}
                           )

        # trigger
        self.add_parameter('trigger_enabled',
                           label='Is trigger enabled',
                           get_cmd=':TRIGger:AND:ENABLe?',
                           set_cmd=':TRIGger:AND:ENABLe {}',
                           val_mapping={True: 1, False: 0}
                           )

        self.add_parameter('trigger_edge_source',
                           label='Source channel for the edge trigger',
                           get_cmd=':TRIGger:EDGE:SOURce?',
                           set_cmd=':TRIGger:EDGE:SOURce {}',
                           vals=Enum(*(
                               [f'CHANnel{i}' for i in range(1, 4 + 1)] +
                               [f'CHAN{i}' for i in range(1, 4 + 1)] +
                               [f'DIGital{i}' for i in range(16 + 1)] +
                               [f'DIG{i}' for i in range(16 + 1)] +
                               ['AUX', 'LINE']))
                           )  # add enum for case insesitivity
        self.add_parameter('trigger_edge_slope',
                           label='slope of the edge trigger',
                           get_cmd=':TRIGger:EDGE:SLOPe?',
                           set_cmd=':TRIGger:EDGE:SLOPe {}',
                           vals=Enum('positive', 'negative', 'neither')
                           )
        self.add_parameter('trigger_level_aux',
                           label='Tirgger level AUX',
                           unit='V',
                           get_cmd=':TRIGger:LEVel? AUX',
                           set_cmd=':TRIGger:LEVel AUX,{}',
                           get_parser=float,
                           vals=Numbers(),
                           )
        # Aquisition
        # If sample points, rate and timebase_scale are set in an
        # incomensurate way, the scope only displays part of the waveform
        self.add_parameter('acquire_points',
                           label='sample points',
                           get_cmd='ACQ:POIN?',
                           get_parser=int,
                           set_cmd=self._cmd_and_invalidate('ACQ:POIN {}'),
                           unit='pts',
                           vals=vals.Numbers(min_value=1, max_value=100e6)
                           )

        self.add_parameter('acquire_sample_rate',
                           label='sample rate',
                           get_cmd='ACQ:SRAT?',
                           set_cmd=self._cmd_and_invalidate('ACQ:SRAT {}'),
                           unit='Sa/s',
                           get_parser=float
                           )

        # this parameter gets used internally for data aquisition. For now it
        # should not be used manually
        self.add_parameter('data_source',
                           label='Waveform Data source',
                           get_cmd=':WAVeform:SOURce?',
                           set_cmd=':WAVeform:SOURce {}',
                           vals = Enum( *(\
                                [f'CHANnel{i}' for i in range(1, 4+1)]+\
                                [f'CHAN{i}' for i in range(1, 4+1)]+\
                                [f'DIFF{i}' for i in range(1, 2+1)]+\
                                [f'COMMonmode{i}' for i in range(3, 4+1)]+\
                                [f'COMM{i}' for i in range(3, 4+1)]+\
                                [f'FUNCtion{i}' for i in range(1, 16+1)]+\
                                [f'FUNC{i}' for i in range(1, 16+1)]+\
                                [f'WMEMory{i}' for i in range(1, 4+1)]+\
                                [f'WMEM{i}' for i in range(1, 4+1)]+\
                                [f'BUS{i}' for i in range(1, 4+1)]+\
                                ['HISTogram', 'HIST', 'CLOCK']+\
                                ['MTRend', 'MTR']))
                           )

        # TODO: implement as array parameter to allow for setting the other filter
        # ratios
        self.add_parameter('acquire_interpolate',
                            get_cmd=':ACQuire:INTerpolate?',
                            set_cmd=self._cmd_and_invalidate(':ACQuire:INTerpolate {}'),
                            val_mapping={True: 1, False: 0}
                            )

        self.add_parameter('acquire_mode',
                            label='Acquisition mode',
                            get_cmd= 'ACQuire:MODE?',
                            set_cmd='ACQuire:MODE {}',
                            vals=Enum('ETIMe', 'RTIMe', 'PDETect',
                                      'HRESolution', 'SEGMented',
                                      'SEGPdetect', 'SEGHres')
                            )

        self.add_parameter('acquire_timespan',
                            get_cmd=(lambda: self.acquire_points.get_latest() \
                                            /self.acquire_sample_rate.get_latest()),
                            unit='s',
                            get_parser=float
                            )

        # time of the first point
        self.add_parameter('waveform_xorigin',
                            get_cmd='WAVeform:XORigin?',
                            unit='s',
                            get_parser=float
                            )

        self.add_parameter('data_format',
                           set_cmd='SAV:WAV:FORM {}',
                           val_mapping={'csv': 'CSV',
                                        'binary': 'BIN',
                                        'asciixy': 'ASC'},
                           docstring=("Set the format for saving "
                                      "files using save_data function")
                           )
        # Channels
        channels = ChannelList(self, "Channels", InfiniiumChannel,
                               snapshotable=False)

        for i in range(1,5):
            channel = InfiniiumChannel(self, f'chan{i}', i)
            channels.append(channel)
            self.add_submodule(f'ch{i}', channel)
        channels.lock()
        self.add_submodule('channels', channels)

        # Submodules
        meassubsys = MeasurementSubsystem(self, 'measure')
        self.add_submodule('measure', meassubsys)

    def _cmd_and_invalidate(self, cmd: str) -> Callable[..., Any]:
        return partial(Infiniium._cmd_and_invalidate_call, self, cmd)

    def _cmd_and_invalidate_call(self, cmd: str, val: float) -> None:
        """
        executes command and sets trace_ready status to false
        any command that effects the number of setpoints should invalidate the trace
        """
        self.trace_ready = False
        self.write(cmd.format(val))

    def save_data(self, filename: str) -> None:
        """
        Saves the channels currently shown on oscilloscope screen to a USB.
        Must set data_format parameter prior to calling this
        """
        self.write(f'SAV:WAV "{filename}"')

    def get_current_traces(
            self,
            channels: Optional[Sequence[int]] = None
    ) -> Dict[Any, Any]:
        """
        Get the current traces of 'channels' on the oscillsocope.

        Args:
            channels: default [1, 2, 3, 4]
                list of integers representing the channels.
                gets the traces of these channels.
                the only valid integers are 1,2,3,4
                will turn any other channels off

        Returns:
            a dict with keys 'ch1', 'ch2', 'ch3', 'ch4', 'time',
            and values are np.ndarrays, corresponding to the voltages
            of the four channels and the common time axis
        """
        if channels is None:
            channels = [1, 2, 3, 4]
        # check that channels are valid
        try:
            assert all([ch in [1, 2, 3, 4] for ch in channels])
        except:
            raise Exception("invalid channel in %s, integers"
                            " must be 1,2,3 or 4" % channels)

        self.write('DIGitize')
        all_data = {}
        self.write(':SYSTem:HEADer OFF')

        for i in channels:
            self.data_source('CHAN%s' % i)
            self.write(':WAVeform:FORMat WORD')
            self.write(":waveform:byteorder LSBFirst")
            self.write(':WAVeform:STReaming OFF')

            data = self.visa_handle.query_binary_values(
                'WAV:DATA?', datatype='h', is_big_endian=False)
            all_data['ch%d' % i] = np.array(data)

        x_incr = float(self.ask(":WAVeform:XINCrement?"))
        y_incr = float(self.ask(":WAVeform:YINCrement?"))
        y_origin = float(self.ask(":WAVeform:YORigin?"))

        for i in channels:
            all_data['ch%d' % i] = all_data['ch%d' % i] * y_incr + y_origin
            self.write(f':CHANnel{i}:DISPlay ON')
        all_data['time'] = np.arange(0, len(all_data['ch%s' % channels[0]])) \
            * x_incr

        self.write(':RUN')
        # turn the channels that were not requested off
        for ch in [i for i in [1, 2, 3, 4] if i not in channels]:
            self.write(f':CHANnel{ch}:DISPlay OFF')

        return all_data



from functools import partial
from typing import Any, Callable, Union

from qcodes import VisaInstrument
from qcodes.utils.validators import Bool, Enum, Ints, MultiType, Numbers


def parse_output_string(s: str) -> str:
    """ Parses and cleans string outputs of the Keithley """
    # Remove surrounding whitespace and newline characters
    s = s.strip()

    # Remove surrounding quotes
    if (s[0] == s[-1]) and s.startswith(("'", '"')):
        s = s[1:-1]

    s = s.lower()

    # Convert some results to a better readable version
    conversions = {
        'mov': 'moving',
        'rep': 'repeat',
    }

    if s in conversions.keys():
        s = conversions[s]

    return s


def parse_output_bool(value: str) -> bool:
    return True if int(value) == 1 else False


class Keithley_2000(VisaInstrument):
    """
    Driver for the Keithley 2000 multimeter.
    """
    def __init__(
            self,
            name: str,
            address: str,
            reset: bool = False,
            **kwargs: Any):
        super().__init__(name, address, terminator='\n', **kwargs)

        self._trigger_sent = False

        # Unfortunately the strings have to contain quotation marks and a
        # newline character, as this is how the instrument returns it.
        self._mode_map = {
            'ac current': '"CURR:AC"',
            'dc current': '"CURR:DC"',
            'ac voltage': '"VOLT:AC"',
            'dc voltage': '"VOLT:DC"',
            '2w resistance': '"RES"',
            '4w resistance': '"FRES"',
            'temperature': '"TEMP"',
            'frequency': '"FREQ"',
        }

        self.add_parameter('mode',
                           get_cmd='SENS:FUNC?',
                           set_cmd="SENS:FUNC {}",
                           val_mapping=self._mode_map)

        # Mode specific parameters
        self.add_parameter('nplc',
                           get_cmd=partial(self._get_mode_param, 'NPLC',
                                           float),
                           set_cmd=partial(self._set_mode_param, 'NPLC'),
                           vals=Numbers(min_value=0.01, max_value=10))

        # TODO: validator, this one is more difficult since different modes
        # require different validation ranges
        self.add_parameter('range',
                           get_cmd=partial(self._get_mode_param, 'RANG',
                                           float),
                           set_cmd=partial(self._set_mode_param, 'RANG'),
                           vals=Numbers())

        self.add_parameter('auto_range_enabled',
                           get_cmd=partial(self._get_mode_param, 'RANG:AUTO',
                                           parse_output_bool),
                           set_cmd=partial(self._set_mode_param, 'RANG:AUTO'),
                           vals=Bool())

        self.add_parameter('digits',
                           get_cmd=partial(self._get_mode_param, 'DIG', int),
                           set_cmd=partial(self._set_mode_param, 'DIG'),
                           vals=Ints(min_value=4, max_value=7))

        self.add_parameter('averaging_type',
                           get_cmd=partial(self._get_mode_param, 'AVER:TCON',
                                           parse_output_string),
                           set_cmd=partial(self._set_mode_param, 'AVER:TCON'),
                           vals=Enum('moving', 'repeat'))

        self.add_parameter('averaging_count',
                           get_cmd=partial(self._get_mode_param, 'AVER:COUN',
                                           int),
                           set_cmd=partial(self._set_mode_param, 'AVER:COUN'),
                           vals=Ints(min_value=1, max_value=100))

        self.add_parameter('averaging_enabled',
                           get_cmd=partial(self._get_mode_param, 'AVER:STAT',
                                           parse_output_bool),
                           set_cmd=partial(self._set_mode_param, 'AVER:STAT'),
                           vals=Bool())

        # Global parameters
        self.add_parameter('display_enabled',
                           get_cmd='DISP:ENAB?',
                           get_parser=parse_output_bool,
                           set_cmd='DISP:ENAB {}',
                           set_parser=int,
                           vals=Bool())

        self.add_parameter('trigger_continuous',
                           get_cmd='INIT:CONT?',
                           get_parser=parse_output_bool,
                           set_cmd='INIT:CONT {}',
                           set_parser=int,
                           vals=Bool())

        self.add_parameter('trigger_count',
                           get_cmd='TRIG:COUN?',
                           get_parser=int,
                           set_cmd='TRIG:COUN {}',
                           vals=MultiType(Ints(min_value=1, max_value=9999),
                                          Enum('inf',
                                               'default',
                                               'minimum',
                                               'maximum')))

        self.add_parameter('trigger_delay',
                           get_cmd='TRIG:DEL?',
                           get_parser=float,
                           set_cmd='TRIG:DEL {}',
                           unit='s',
                           vals=Numbers(min_value=0, max_value=999999.999))

        self.add_parameter('trigger_source',
                           get_cmd='TRIG:SOUR?',
                           set_cmd='TRIG:SOUR {}',
                           val_mapping={
                               'immediate': 'IMM',
                               'timer': 'TIM',
                               'manual': 'MAN',
                               'bus': 'BUS',
                               'external': 'EXT',
                           })

        self.add_parameter('trigger_timer',
                           get_cmd='TRIG:TIM?',
                           get_parser=float,
                           set_cmd='TRIG:TIM {}',
                           unit='s',
                           vals=Numbers(min_value=0.001, max_value=999999.999))

        self.add_parameter('amplitude',
                           unit='arb.unit',
                           get_cmd=self._read_next_value)

        self.add_function('reset', call_cmd='*RST')

        if reset:
            self.reset()

        # Set the data format to have only ascii data without units and channels
        self.write('FORM:DATA ASCII')
        self.write('FORM:ELEM READ')

        self.connect_message()

    def trigger(self) -> None:
        if not self.trigger_continuous():
            self.write('INIT')
            self._trigger_sent = True

    def _read_next_value(self) -> float:
        # Prevent a timeout when no trigger has been sent
        if not self.trigger_continuous() and not self._trigger_sent:
            return 0.0

        self._trigger_sent = False

        return float(self.ask('SENSE:DATA:FRESH?'))

    def _get_mode_param(
            self,
            parameter: str,
            parser: Callable[[str], Any]
    ) -> Union[float, str, bool]:
        """ Read the current Keithley mode and ask for a parameter """
        mode = parse_output_string(self._mode_map[self.mode()])
        cmd = f'{mode}:{parameter}?'

        return parser(self.ask(cmd))

    def _set_mode_param(self, parameter: str, value: Union[float, str, bool]) -> None:
        """ Read the current Keithley mode and set a parameter """
        if isinstance(value, bool):
            value = int(value)

        mode = parse_output_string(self._mode_map[self.mode()])
        cmd = f'{mode}:{parameter} {value}'

        self.write(cmd)



import pytest
from typing import Dict, Callable, Any
import logging
from functools import wraps
from contextlib import suppress
import time

from qcodes.instrument.base import InstrumentBase
from qcodes.logger.instrument_logger import get_instrument_logger
from qcodes.instrument_drivers.Lakeshore.Model_372 import Model_372
import qcodes.instrument.sims as sims
from qcodes.instrument_drivers.Lakeshore.lakeshore_base import BaseSensorChannel

log = logging.getLogger(__name__)

VISA_LOGGER = '.'.join((InstrumentBase.__module__, 'com', 'visa'))


class MockVisaInstrument:
    """
    Mixin class that overrides write_raw and ask_raw to simulate an
    instrument.
    """
    def __init__(self, *args, **kwargs) -> None:
        # ignore this line in mypy: Mypy does not support mixins yet
        # and seen by itself with this class definition it does not make sense
        # to call __init__ on the super()
        super().__init__(*args, **kwargs)  # type: ignore[call-arg]
        self.visa_log = get_instrument_logger(self, VISA_LOGGER)

        # This base class mixin holds two dictionaries associated with the
        # pyvisa_instrument.write()
        self.cmds: Dict[str, Callable[..., Any]] = {}
        # and pyvisa_instrument.query() functions
        self.queries: Dict[str, Callable[..., Any]] = {}
        # the keys are the issued VISA commands like '*IDN?' or '*OPC'
        # the values are the corresponding methods to be called on the mock
        # instrument.

        # To facilitate the definition there are the decorators `@query' and
        # `@command`. These attach an attribute to the method, so that the
        # dictionaries can be filled here in the constructor. (This is
        # borderline abusive, but makes a it easy to define mocks)
        func_names = dir(self)
        # cycle through all methods
        for func_name in func_names:
            f = getattr(self, func_name)
            # only add for methods that have such an attribute
            with suppress(AttributeError):
                self.queries[getattr(f, 'query_name')] = f
            with suppress(AttributeError):
                self.cmds[getattr(f, 'command_name')] = f

    def write_raw(self, cmd):
        cmd_parts = cmd.split(' ')
        cmd_str = cmd_parts[0].upper()
        if cmd_str in self.cmds:
            args = ''.join(cmd_parts[1:])
            self.visa_log.debug(f'Query: '
                      f'{cmd} for command {cmd_str} with args {args}')
            self.cmds[cmd_str](args)
        else:
            super().write_raw(cmd)

    def ask_raw(self, query):
        query_parts = query.split(' ')
        query_str = query_parts[0].upper()
        if query_str in self.queries:
            args = ''.join(query_parts[1:])
            self.visa_log.debug(f'Query: '
                      f'{query} for command {query_str} with args {args}')
            response = self.queries[query_str](args)
            self.visa_log.debug(f"Response: {response}")
            return response
        else:
            super().ask_raw(query)


def query(name=None):
    def wrapper(func):
        func.query_name = name.upper()
        return func
    return wrapper


def command(name=None):
    def wrapper(func):
        func.command_name = name.upper()
        return func
    return wrapper


def split_args(split_char=','):
    def wrapper(func):
        @wraps(func)
        def decorated_func(self, string_arg):
            args = string_arg.split(split_char)
            return func(self, *args)
        return decorated_func
    return wrapper


class DictClass:
    def __init__(self, **kwargs):
        for kwarg, value in kwargs.items():
            setattr(self, kwarg, value)


class Model_372_Mock(MockVisaInstrument, Model_372):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # initial values
        self.heaters: Dict[str, DictClass] = {}
        self.heaters['0'] = DictClass(P=1, I=2, D=3,
                                      mode=5, input_channel=2,
                                      powerup_enable=0, polarity=0,
                                      use_filter=0, delay=1,
                                      output_range=0,
                                      setpoint=4)
        self.heaters['1'] = DictClass(P=1, I=2, D=3,
                                      mode=5, input_channel=2,
                                      powerup_enable=0, polarity=0,
                                      use_filter=0, delay=1,
                                      output_range=0,
                                      setpoint=4)
        self.heaters['2'] = DictClass(P=1, I=2, D=3,
                                      mode=5, input_channel=2,
                                      powerup_enable=0, polarity=0,
                                      use_filter=0, delay=1,
                                      output_range=0,
                                      setpoint=4)

        self.channel_mock = \
            {str(i): DictClass(tlimit=i, T=4, enabled=1, # True
                               dwell=100, pause=3,
                               curve_number=0,
                               temperature_coefficient=1, # 'negative',
                               excitation_mode=0, #'voltage',
                               excitation_range_number=1,
                               auto_range=0,#'off',
                               range=5,#'200 mOhm',
                               current_source_shunted=0,#False,
                               units=1)#'kelvin')
             for i in range(1, 17)}

        # simulate delayed heating
        self.simulate_heating = False
        self.start_heating_time = time.perf_counter()

    def start_heating(self):
        self.start_heating_time = time.perf_counter()
        self.simulate_heating = True

    def get_t_when_heating(self):
        """
        Simply define a fixed setpoint of 4 k for now
        """
        delta = abs(time.perf_counter() - self.start_heating_time)
        # make it simple to start with: linear ramp 1K per second
        # start at 7K.
        return max(4, 7 - delta)

    @query('PID?')
    def pidq(self, arg):
        heater = self.heaters[arg]
        return f'{heater.P},{heater.I},{heater.D}'

    @command('PID')
    @split_args()
    def pid(self, output, P, I, D):
        for a, v in zip(['P', 'I', 'D'], [P, I, D]):
            setattr(self.heaters[output], a, v)

    @query('OUTMODE?')
    def outmodeq(self, arg):
        heater = self.heaters[arg]
        return (f'{heater.mode},{heater.input_channel},'
                f'{heater.powerup_enable},{heater.polarity},'
                f'{heater.use_filter},{heater.delay}')

    @command('OUTMODE')
    @split_args()
    def outputmode(self, output, mode, input_channel, powerup_enable,
                   polarity, use_filter, delay):
        h = self.heaters[output]
        h.output = output
        h.mode = mode
        h.input_channel = input_channel
        h.powerup_enable = powerup_enable
        h.polarity = polarity
        h.use_filter = use_filter
        h.delay = delay

    @query('INSET?')
    def insetq(self, channel):
        ch = self.channel_mock[channel]
        return (f'{ch.enabled},{ch.dwell},'
                f'{ch.pause},{ch.curve_number},'
                f'{ch.temperature_coefficient}')

    @command('INSET')
    @split_args()
    def inset(self, channel, enabled, dwell, pause, curve_number,
              temperature_coefficient):
        ch = self.channel_mock[channel]
        ch.enabled = enabled
        ch.dwell = dwell
        ch.pause = pause
        ch.curve_number = curve_number
        ch.temperature_coefficient = temperature_coefficient

    @query('INTYPE?')
    def intypeq(self, channel):
        ch = self.channel_mock[channel]
        return (f'{ch.excitation_mode},{ch.excitation_range_number},'
                f'{ch.auto_range},{ch.range},'
                f'{ch.current_source_shunted},{ch.units}')

    @command('INTYPE')
    @split_args()
    def intype(self, channel, excitation_mode, excitation_range_number,
               auto_range, range, current_source_shunted, units):
        ch = self.channel_mock[channel]
        ch.excitation_mode = excitation_mode
        ch.excitation_range_number = excitation_range_number
        ch.auto_range = auto_range
        ch.range = range
        ch.current_source_shunted = current_source_shunted
        ch.units = units

    @query('RANGE?')
    def rangeq(self, heater):
        h = self.heaters[heater]
        return f'{h.output_range}'

    @command('RANGE')
    @split_args()
    def range_cmd(self, heater, output_range):
        h = self.heaters[heater]
        h.output_range = output_range

    @query('SETP?')
    def setpointq(self, heater):
        h = self.heaters[heater]
        return f'{h.setpoint}'

    @command('SETP')
    @split_args()
    def setpoint(self, heater, setpoint):
        h = self.heaters[heater]
        h.setpoint = setpoint

    @query('TLIMIT?')
    def tlimitq(self, channel):
        chan = self.channel_mock[channel]
        return f'{chan.tlimit}'

    @command('TLIMIT')
    @split_args()
    def tlimitcmd(self, channel, tlimit):
        chan = self.channel_mock[channel]
        chan.tlimit = tlimit

    @query('KRDG?')
    def temperature(self, output):
        chan = self.channel_mock[output]
        if self.simulate_heating:
            return self.get_t_when_heating()
        return f'{chan.T}'


def instrument_fixture(scope='function'):
    def wrapper(func):
        @pytest.fixture(scope=scope)
        def wrapped_fixture():
            inst = func()
            try:
                yield inst
            finally:
                inst.close()
        return wrapped_fixture
    return wrapper


@instrument_fixture(scope='function')
def lakeshore_372():
    visalib = sims.__file__.replace('__init__.py',
                                    'lakeshore_model372.yaml@sim')
    return Model_372_Mock('lakeshore_372_fixture', 'GPIB::3::INSTR',
                          visalib=visalib, device_clear=False)


def test_pid_set(lakeshore_372):
    ls = lakeshore_372
    P, I, D = 1, 2, 3
    for h in (ls.warmup_heater, ls.analog_heater, ls.sample_heater):
        h.P(P)
        h.I(I)
        h.D(D)
        assert (h.P(), h.I(), h.D()) == (P, I, D)


def test_output_mode(lakeshore_372):
    ls = lakeshore_372
    mode = 'off'
    input_channel = 1
    powerup_enable = True
    polarity = 'unipolar'
    use_filter = True
    delay = 1
    for h in (ls.warmup_heater, ls.analog_heater, ls.sample_heater):
        h.mode(mode)
        h.input_channel(input_channel)
        h.powerup_enable(powerup_enable)
        h.polarity(polarity)
        h.use_filter(use_filter)
        h.delay(delay)
        assert h.mode() == mode
        assert h.input_channel() == input_channel
        assert h.powerup_enable() == powerup_enable
        assert h.polarity() == polarity
        assert h.use_filter() == use_filter
        assert h.delay() == delay


def test_range(lakeshore_372):
    ls = lakeshore_372
    output_range = '10mA'
    for h in (ls.warmup_heater, ls.analog_heater, ls.sample_heater):
        h.output_range(output_range)
        assert h.output_range() == output_range


def test_tlimit(lakeshore_372):
    ls = lakeshore_372
    tlimit = 5.1
    for ch in ls.channels:
        ch.t_limit(tlimit)
        assert ch.t_limit() == tlimit


def test_setpoint(lakeshore_372):
    ls = lakeshore_372
    setpoint = 5.1
    for h in (ls.warmup_heater, ls.analog_heater, ls.sample_heater):
        h.setpoint(setpoint)
        assert h.setpoint() == setpoint


def test_select_range_limits(lakeshore_372):
    h = lakeshore_372.sample_heater
    ranges = list(range(1, 9))
    h.range_limits(ranges)

    for i in ranges:
        h.set_range_from_temperature(i - 0.5)
        assert h.output_range() == h.INVERSE_RANGES[i]

    h.set_range_from_temperature(i + 0.5)
    assert h.output_range() == h.INVERSE_RANGES[len(ranges)]


def test_set_and_wait_unit_setpoint_reached(lakeshore_372):
    ls = lakeshore_372
    ls.sample_heater.setpoint(4)
    ls.start_heating()
    ls.sample_heater.wait_until_set_point_reached()


def test_blocking_t(lakeshore_372):
    ls = lakeshore_372
    h = lakeshore_372.sample_heater
    ranges = list(range(1, 9))
    h.range_limits(ranges)
    ls.start_heating()
    h.blocking_t(4)


def test_get_term_sum():
    available_terms = [0, 1, 2, 4, 8, 16, 32]

    assert [32, 8, 2, 1] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            1 + 2 + 8 + 32)

    assert [32] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            32)

    assert [16, 4, 1] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            1 + 4 + 16)

    assert [0] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            0)


def test_get_term_sum_with_some_powers_of_2_omitted():
    available_terms = [0, 16, 32]

    assert [32, 16] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            16 + 32)

    assert [32] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            32)

    assert [0] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            0)


def test_get_term_sum_returns_empty_list():
    available_terms = [0, 16, 32]

    assert [] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            15)

def test_get_term_sum_when_zero_is_not_in_available_terms():
    available_terms = [16, 32]

    assert [] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            3)

    # Note that `_get_sum_terms` expects '0' to be in the available_terms,
    # hence for this particular case it will still return a list with '0' in
    # it although that '0' is not part of the available_terms
    assert [0] == \
           BaseSensorChannel._get_sum_terms(available_terms,
                                            0)



from typing import ClassVar, Dict, Any

from qcodes.instrument.group_parameter import GroupParameter, Group
from .lakeshore_base import LakeshoreBase, BaseOutput, BaseSensorChannel
import qcodes.utils.validators as vals


# There are 4 sensors channels (a.k.a. measurement inputs) in Model 336.
# Unlike other Lakeshore models, Model 336 refers to the channels using
# letters, and not numbers
_channel_name_to_command_map: Dict[str, str] = {'A': 'A',
                                                'B': 'B',
                                                'C': 'C',
                                                'D': 'D'}

# OUTMODE command of this model refers to the outputs via integer numbers,
# while everywhere else within this model letters are used. This map is
# created in order to preserve uniformity of referencing to sensor channels
# within this driver.
_channel_name_to_outmode_command_map: Dict[str, int] = \
    {ch_name: num_for_cmd + 1
     for num_for_cmd, ch_name in enumerate(_channel_name_to_command_map.keys())}


class Output_336_CurrentSource(BaseOutput):
    """
    Class for control outputs 1 and 2 of model 336 that are variable DC current
    sources referenced to chassis ground
    """

    MODES: ClassVar[Dict[str, int]] = {
        'off': 0,
        'closed_loop': 1,
        'zone': 2,
        'open_loop': 3}

    RANGES: ClassVar[Dict[str, int]] = {
        'off': 0,
        'low': 1,
        'medium': 2,
        'high': 3}

    _input_channel_parameter_kwargs = {
        'val_mapping': _channel_name_to_outmode_command_map}

    def __init__(
            self,
            parent: "Model_336",
            output_name: str,
            output_index: int
    ):
        super().__init__(parent, output_name, output_index, has_pid=True)

        self.P.vals = vals.Numbers(0.1, 1000)
        self.I.vals = vals.Numbers(0.1, 1000)
        self.D.vals = vals.Numbers(0, 200)


class Output_336_VoltageSource(BaseOutput):
    """
    Class for control outputs 3 and 4 of model 336 that are variable DC voltage
    sources
    """

    MODES: ClassVar[Dict[str, int]] = {
        'off': 0,
        'closed_loop': 1,
        'zone': 2,
        'open_loop': 3,
        'monitor_out': 4,
        'warm_up': 5}

    RANGES: ClassVar[Dict[str, int]] = {
        'off': 0,
        'low': 1,
        'medium': 2,
        'high': 3}

    def __init__(
            self,
            parent: "Model_336",
            output_name: str,
            output_index: int
    ):
        super().__init__(parent, output_name, output_index, has_pid=False)


class Model_336_Channel(BaseSensorChannel):
    SENSOR_STATUSES = {0: 'OK',
                       1: 'Invalid Reading',
                       16: 'Temp Underrange',
                       32: 'Temp Overrange',
                       64: 'Sensor Units Zero',
                       128: 'Sensor Units Overrange'}

    def __init__(
            self,
            parent: "Model_336",
            name: str,
            channel: str
    ):
        super().__init__(parent, name, channel)

        # Parameters related to Input Type Parameter Command (INTYPE)
        self.add_parameter('sensor_type',
                           label='Input sensor type',
                           docstring='Specifies input sensor type',
                           val_mapping={'disabled': 0,
                                        'diode': 1,
                                        'platinum_rtd': 2,
                                        'ntc_rtd': 3},
                           parameter_class=GroupParameter)
        self.add_parameter('auto_range_enabled',
                           label='Autoranging',
                           docstring='Specifies if autoranging is enabled. '
                                     'Does not apply for diode sensor type',
                           val_mapping={False: 0, True: 1},
                           parameter_class=GroupParameter)
        self.add_parameter('range',
                           label='Range',
                           docstring='Specifies input range when autorange is '
                                     'not enabled. If autorange is on, the '
                                     'returned value corresponds to the '
                                     'currently auto-selected range. The list '
                                     'of available ranges depends on the '
                                     'chosen sensor type: diode 0-1, platinum '
                                     'RTD 0-6, NTC RTD 0-8. Refer to the page '
                                     '136 of the manual for the lookup table',
                           vals=vals.Numbers(0, 8),
                           parameter_class=GroupParameter)
        self.add_parameter('compensation_enabled',
                           label='Compensation enabled',
                           docstring='Specifies input compensation. Reversal '
                                     'for thermal EMF compensation if input '
                                     'is resistive, room compensation if '
                                     'input is thermocouple. Always 0 if input '
                                     'is a diode',
                           val_mapping={False: 0, True: 1},
                           parameter_class=GroupParameter)
        self.add_parameter('units',
                           label='Preferred units',
                           docstring='Specifies the preferred units parameter '
                                     'for sensor readings and for the control '
                                     'setpoint (kelvin, celsius, or sensor)',
                           val_mapping={'kelvin': 1, 'celsius': 2, 'sensor': 3},
                           parameter_class=GroupParameter)
        self.output_group = Group([self.sensor_type, self.auto_range_enabled,
                                   self.range, self.compensation_enabled,
                                   self.units],
                                  set_cmd=f'INTYPE {self._channel}, '
                                          f'{{sensor_type}}, '
                                          f'{{auto_range_enabled}}, {{range}}, '
                                          f'{{compensation_enabled}}, '
                                          f'{{units}}',
                                  get_cmd=f'INTYPE? {self._channel}')


class Model_336(LakeshoreBase):
    """
    Lakeshore Model 336 Temperature Controller Driver
    """
    channel_name_command: Dict[str, str] = _channel_name_to_command_map

    CHANNEL_CLASS = Model_336_Channel

    input_channel_parameter_values_to_channel_name_on_instrument = \
        _channel_name_to_command_map

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        super().__init__(name, address, **kwargs)

        self.output_1 = Output_336_CurrentSource(self, 'output_1', 1)
        self.output_2 = Output_336_CurrentSource(self, 'output_2', 2)
        self.output_3 = Output_336_VoltageSource(self, 'output_3', 3)
        self.output_4 = Output_336_VoltageSource(self, 'output_4', 4)



"""
This module provides a wrapper class :class:`ConnectionPlus` around
:class:`sqlite3.Connection` together with functions around it which allow
performing nested atomic transactions on an SQLite database.
"""
import logging
import sqlite3
from contextlib import contextmanager
from typing import Union, Any, Iterator

import wrapt

from qcodes.utils.delaykeyboardinterrupt import DelayedKeyboardInterrupt

log = logging.getLogger(__name__)


class ConnectionPlus(wrapt.ObjectProxy):
    """
    A class to extend the sqlite3.Connection object. Since sqlite3.Connection
    has no __dict__, we can not directly add attributes to its instance
    directly.

    It is not allowed to instantiate a new `ConnectionPlus` object from a
    `ConnectionPlus` object.

    Attributes:
        atomic_in_progress: a bool describing whether the connection is
            currently in the middle of an atomic block of transactions, thus
            allowing to nest `atomic` context managers
        path_to_dbfile: Path to the database file of the connection.
    """
    atomic_in_progress: bool = False
    path_to_dbfile = ''

    def __init__(self, sqlite3_connection: sqlite3.Connection):
        super().__init__(sqlite3_connection)

        if isinstance(sqlite3_connection, ConnectionPlus):
            raise ValueError('Attempted to create `ConnectionPlus` from a '
                             '`ConnectionPlus` object which is not allowed.')

        self.path_to_dbfile = path_to_dbfile(sqlite3_connection)


def make_connection_plus_from(conn: Union[sqlite3.Connection, ConnectionPlus]
                              ) -> ConnectionPlus:
    """
    Makes a ConnectionPlus connection object out of a given argument.

    If the given connection is already a ConnectionPlus, then it is returned
    without any changes.

    Args:
        conn: an sqlite database connection object

    Returns:
        the "same" connection but as ConnectionPlus object
    """
    if not isinstance(conn, ConnectionPlus):
        conn_plus = ConnectionPlus(conn)
    else:
        conn_plus = conn
    return conn_plus


@contextmanager
def atomic(conn: ConnectionPlus) -> Iterator[ConnectionPlus]:
    """
    Guard a series of transactions as atomic.

    If one transaction fails, all the previous transactions are rolled back
    and no more transactions are performed.

    NB: 'BEGIN' is by default only inserted before INSERT/UPDATE/DELETE/REPLACE
    but we want to guard any transaction that modifies the database (e.g. also
    ALTER)

    Args:
        conn: connection to guard
    """
    with DelayedKeyboardInterrupt():
        if not isinstance(conn, ConnectionPlus):
            raise ValueError('atomic context manager only accepts '
                             'ConnectionPlus database connection objects.')

        is_outmost = not(conn.atomic_in_progress)

        if conn.in_transaction and is_outmost:
            raise RuntimeError('SQLite connection has uncommitted '
                               'transactions. '
                               'Please commit those before starting an atomic '
                               'transaction.')

        old_atomic_in_progress = conn.atomic_in_progress
        conn.atomic_in_progress = True

        try:
            if is_outmost:
                old_level = conn.isolation_level
                conn.isolation_level = None
                conn.cursor().execute('BEGIN')
            yield conn
        except Exception as e:
            conn.rollback()
            log.exception("Rolling back due to unhandled exception")
            raise RuntimeError("Rolling back due to unhandled exception") from e
        else:
            if is_outmost:
                conn.commit()
        finally:
            if is_outmost:
                conn.isolation_level = old_level
            conn.atomic_in_progress = old_atomic_in_progress


def transaction(conn: ConnectionPlus,
                sql: str, *args: Any) -> sqlite3.Cursor:
    """Perform a transaction.
    The transaction needs to be committed or rolled back.


    Args:
        conn: database connection
        sql: formatted string
        *args: arguments to use for parameter substitution

    Returns:
        sqlite cursor

    """
    c = conn.cursor()
    if len(args) > 0:
        c.execute(sql, args)
    else:
        c.execute(sql)
    return c


def atomic_transaction(conn: ConnectionPlus,
                       sql: str, *args: Any) -> sqlite3.Cursor:
    """Perform an **atomic** transaction.
    The transaction is committed if there are no exceptions else the
    transaction is rolled back.
    NB: 'BEGIN' is by default only inserted before INSERT/UPDATE/DELETE/REPLACE
    but we want to guard any transaction that modifies the database (e.g. also
    ALTER). 'BEGIN' marks a place to commit from/roll back to

    Args:
        conn: database connection
        sql: formatted string
        *args: arguments to use for parameter substitution

    Returns:
        sqlite cursor

    """
    with atomic(conn) as atomic_conn:
        c = transaction(atomic_conn, sql, *args)
    return c


def path_to_dbfile(conn: Union[ConnectionPlus, sqlite3.Connection]) -> str:
    """
    Return the path of the database file that the conn object is connected to
    """
    # according to https://www.sqlite.org/pragma.html#pragma_database_list
    # the 3th element (1 indexed) is the path
    cursor = conn.cursor()
    cursor.execute("PRAGMA database_list")
    row = cursor.fetchall()[0]

    return row[2]



"""
This module defines the Link dataclass as well as two functions to read and
write an Link object to/from string, respectively
"""
from typing import List
from dataclasses import dataclass, asdict
import json
from qcodes.dataset.guids import validate_guid_format


@dataclass
class Link:
    """
    Class to represent a link between two datasets. The link is a little graph
    with two nodes (head and tail) and a directed edge.

    Attributes:
        head: a guid representing the head of the graph
        tail: a guid representing the tail of the graph
        edge_type: a name to represent the type of the edge
        description: free-form optional field add a description of the graph
    """
    head: str
    tail: str
    edge_type: str
    description: str = ""

    @staticmethod
    def validate_node(node_guid: str, node: str) -> None:
        """
        Validate that the guid given is a valid guid.

        Args:
            node_guid: the guid
            node: either "head" or "tail"
        """
        try:
            validate_guid_format(node_guid)
        except ValueError:
            raise ValueError(
                f'The guid given for {node} is not a valid guid. Received '
                f'{node_guid}.')

    def __post_init__(self) -> None:
        self.validate_node(self.head, "head")
        self.validate_node(self.tail, "tail")


def link_to_str(link: Link) -> str:
    """
    Convert a Link to a string
    """
    return json.dumps(asdict(link))


def str_to_link(string: str) -> Link:
    """
    Convert a string to a Link
    """
    ldict = json.loads(string)
    link = Link(**ldict)
    return link


def links_to_str(links: List[Link]) -> str:
    """
    Convert a list of links to string. Note that this is the output that gets
    stored in the DB file
    """
    output = json.dumps([link_to_str(link) for link in links])
    return output


def str_to_links(links_string: str) -> List[Link]:
    """
    Convert a string into a list of Links
    """
    if links_string == '[]':
        return []
    link_dicts = [json.loads(l_str) for l_str in json.loads(links_string)]
    links = [Link(**ld) for ld in link_dicts]
    return links



from typing import List

import numpy as np
import pytest
from hypothesis import HealthCheck, assume, example, given, settings
from hypothesis.strategies import data, floats, just, lists, one_of, sampled_from, text
from matplotlib.collections import QuadMesh

import qcodes as qc
from qcodes.dataset.data_export import DSPlotData
from qcodes.dataset.descriptions.detect_shapes import detect_shape_of_measurement
from qcodes.dataset.measurements import Measurement
from qcodes.dataset.plotting import (
    _appropriate_kwargs,
    _complex_to_real_preparser,
    _make_rescaled_ticks_and_units,
    plot_by_id,
    plot_dataset,
)
from qcodes.tests.instrument_mocks import DummyInstrument
from qcodes.utils.plotting import _ENGINEERING_PREFIXES, _UNITS_FOR_RESCALING


class TerminateLoopException(Exception):
    pass


@given(
    param_name=text(min_size=1, max_size=10),
    param_label=text(min_size=0, max_size=15),
    scale=sampled_from(sorted(list(_ENGINEERING_PREFIXES.keys()))),
    unit=sampled_from(
        sorted(
            list(_UNITS_FOR_RESCALING.union(["", "unit", "kg", "%", "permille", "nW"]))
        )
    ),
    data_strategy=data(),
)
@example(
    param_name="huge_param",
    param_label="Larger than the highest scale",
    scale=max(list(_ENGINEERING_PREFIXES.keys())),
    unit="V",
    data_strategy=np.random.random((5,))
    * 10 ** (3 + max(list(_ENGINEERING_PREFIXES.keys()))),
)
@example(
    param_name="small_param",
    param_label="Lower than the lowest scale",
    scale=min(list(_ENGINEERING_PREFIXES.keys())),
    unit="V",
    data_strategy=np.random.random((5,))
    * 10 ** (-3 + min(list(_ENGINEERING_PREFIXES.keys()))),
)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_rescaled_ticks_and_units(scale, unit,
                                  param_name, param_label, data_strategy):
    if isinstance(data_strategy, np.ndarray):
        # No need to generate data, because it is being passed
        data_array = data_strategy
    else:
        # Generate actual data
        scale_upper_bound = 999.9999 * 10 ** scale
        scale_lower_bound = 10 ** scale
        data_array = np.array(data_strategy.draw(
                    lists(elements=one_of(
                        floats(max_value=scale_upper_bound,
                               min_value=-scale_upper_bound),
                        just(np.nan)
                    ), min_size=1)))

        data_array_not_nans = data_array[np.logical_not(np.isnan(data_array))]

        # data_array with nan values *only* is not supported
        assume(len(data_array_not_nans) > 0)

        # data_array has to contain at least one number within the scale of
        # interest (meaning absolute value of at least one number has to
        # be larger than the lowest value within the scale of interest)
        assume((scale_lower_bound < np.abs(data_array_not_nans)).any())

    data_dict = {
        'name': param_name,
        'label': param_label,
        'unit': unit,
        'data': data_array
    }

    ticks_formatter, label = _make_rescaled_ticks_and_units(data_dict)
    if unit in _UNITS_FOR_RESCALING:
        expected_prefix = _ENGINEERING_PREFIXES[scale]
    else:
        if scale != 0 and unit != "":
            expected_prefix = f'$10^{{{scale:.0f}}}$ '
        else:
            expected_prefix = ''
    if param_label == '':
        base_label = param_name
    else:
        base_label = param_label
    postfix = expected_prefix + unit
    if postfix != '':
        assert f"{base_label} ({postfix})" == label
    else:
        assert f"{base_label}" == label

    if unit != '':
        assert '5' == ticks_formatter(5 / (10 ** (-scale)))
        assert '1' == ticks_formatter(1 / (10 ** (-scale)))
        # also test the fact that "{:g}" is used in ticks formatter function
        assert '2.12346' == ticks_formatter(2.123456789 / (10 ** (-scale)))


def test_plot_by_id_line_and_heatmap(experiment, request):
    """
    Test that line plots and heatmaps can be plotted together
    """
    inst = DummyInstrument('dummy', gates=['s1', 'm1', 's2', 'm2'])
    request.addfinalizer(inst.close)

    inst.m1.get = np.random.randn
    inst.m2.get = lambda: np.random.randint(0, 5)

    meas = Measurement()
    meas.register_parameter(inst.s1)
    meas.register_parameter(inst.s2)
    meas.register_parameter(inst.m2, setpoints=(inst.s1, inst.s2))
    meas.register_parameter(inst.m1, setpoints=(inst.s1,))

    with meas.run() as datasaver:
        for outer in range(10):
            datasaver.add_result((inst.s1, outer), (inst.m1, inst.m1()))
            for inner in range(10):
                datasaver.add_result(
                    (inst.s1, outer), (inst.s2, inner), (inst.m2, inst.m2())
                )

    dataid = datasaver.run_id
    plot_by_id(dataid)
    plot_by_id(dataid, cmap='bone')


@pytest.mark.parametrize("nan_setpoints", [True, False])
@pytest.mark.parametrize("shifted", [True, False])
def test_plot_dataset_2d_shaped(experiment, request, nan_setpoints, shifted):
    """
    Test plotting of preshaped data on a grid that may or may not be shifted
    with and without nans in the set points.
    """
    inst = DummyInstrument("dummy", gates=["s1", "m1", "s2"])
    request.addfinalizer(inst.close)

    inst.m1.get = np.random.randn

    meas = Measurement()
    meas.register_parameter(inst.s1)
    meas.register_parameter(inst.s2)
    meas.register_parameter(inst.m1, setpoints=(inst.s1, inst.s2))

    outer_shape = 10
    inner_shape = 20

    meas.set_shapes(detect_shape_of_measurement((inst.m1,), (outer_shape, inner_shape)))

    shift = 0

    with meas.run() as datasaver:
        try:
            for outer in np.linspace(0, 9, outer_shape):
                for inner in np.linspace(0 + shift, 10 + shift, inner_shape):
                    datasaver.add_result(
                        (inst.s1, outer), (inst.s2, inner), (inst.m1, inst.m1())
                    )
                    if inner > 7 and outer > 6 and nan_setpoints:
                        raise TerminateLoopException
                if shifted:
                    shift += 1
        except TerminateLoopException:
            pass

    axes, cbs = plot_dataset(datasaver.dataset)
    xlims = axes[0].get_xlim()
    ylims = axes[0].get_ylim()

    # check that this generates a QuadMesh which is the expected output of pcolormesh
    assert any(isinstance(mplobj, QuadMesh) for mplobj in axes[0].get_children())

    if nan_setpoints and shifted:
        assert xlims[0] == -0.5
        assert xlims[1] == 7.5
        assert ylims[0] < 0
        assert ylims[0] > -1.0
        assert ylims[1] > 16
        assert ylims[1] < 17
    elif not nan_setpoints and shifted:
        assert xlims[0] == -0.5
        assert xlims[1] == 9.5
        assert ylims[0] < 0
        assert ylims[0] > -1.0
        assert ylims[1] > 19
        assert ylims[1] < 20
    elif nan_setpoints and not shifted:
        assert xlims[0] == -0.5
        assert xlims[1] == 7.5
        assert ylims[0] < 0
        assert ylims[0] > -1.0
        assert ylims[1] > 10
        assert ylims[1] < 11
    else:
        assert xlims[0] == -0.5
        assert xlims[1] == 9.5
        assert ylims[0] < 0
        assert ylims[0] > -1.0
        assert ylims[1] > 10
        assert ylims[1] < 11


def test_appropriate_kwargs():

    kwargs = {'cmap': 'bone'}
    check = kwargs.copy()

    with _appropriate_kwargs('1D_line', False, **kwargs) as ap_kwargs:
        assert ap_kwargs == {}

    assert kwargs == check

    with _appropriate_kwargs('1D_point', False, **kwargs) as ap_kwargs:
        assert ap_kwargs == {}

    assert kwargs == check

    with _appropriate_kwargs('1D_bar', False, **kwargs) as ap_kwargs:
        assert ap_kwargs == {}

    assert kwargs == check

    with _appropriate_kwargs('2D_grid', False, **kwargs) as ap_kwargs:
        assert ap_kwargs == kwargs

    assert kwargs == check

    with _appropriate_kwargs('2D_point', False, **{}) as ap_kwargs:
        assert len(ap_kwargs) == 1
        assert ap_kwargs['cmap'] == qc.config.plotting.default_color_map


def test__complex_to_real_preparser_complex_toplevel_param() -> None:

    data_in: List[List[DSPlotData]] = [
        [
            {
                "data": np.array([0, 1, 2]),
                "name": "voltage",
                "label": "swept voltage",
                "unit": "V",
                "shape": None,
            },
            {
                "data": np.array([0 + 0j, 1 + 2j, -1 + 1j]),
                "name": "signal",
                "label": "complex signal",
                "unit": "Ohm",
                "shape": None,
            },
        ]
    ]

    data_out = _complex_to_real_preparser(data_in, conversion='real_and_imag')

    assert np.shape(np.array(data_out)) == (2, 2)
    assert data_out[0][0] == data_in[0][0]

    real_param = data_out[0][1]
    assert real_param["name"] == "signal_real"
    assert real_param["label"] == "complex signal [real]"
    assert all(real_param["data"] == np.array([0, 1, -1]))
    assert real_param["unit"] == "Ohm"

    imag_param = data_out[1][1]
    assert imag_param['name'] == 'signal_imag'
    assert imag_param['label'] == 'complex signal [imag]'
    assert all(imag_param['data'] == np.array([0, 2, 1]))
    assert imag_param['unit'] == 'Ohm'

    data_out = _complex_to_real_preparser(data_in, conversion='mag_and_phase')

    assert len(data_out) == 2
    assert len(data_out[0]) == 2
    assert data_out[0][0] == data_in[0][0]

    phase_param = data_out[1][1]
    assert phase_param["name"] == "signal_phase"
    assert phase_param["label"] == "complex signal [phase]"
    assert all(phase_param["data"] == np.angle(np.array([0 + 0j, 1 + 2j, -1 + 1j])))
    assert phase_param["unit"] == "rad"

    mag_param = data_out[0][1]
    assert mag_param['name'] == 'signal_mag'
    assert mag_param['label'] == 'complex signal [mag]'
    assert all(mag_param['data'] == np.array([0, np.sqrt(5), np.sqrt(2)]))
    assert mag_param['unit'] == 'Ohm'

    data_out = _complex_to_real_preparser(data_in, conversion='mag_and_phase',
                                          degrees=True)

    phase_param = data_out[1][1]
    assert phase_param["name"] == "signal_phase"
    assert phase_param["label"] == "complex signal [phase]"
    assert all(
        phase_param["data"] == np.angle(np.array([0 + 0j, 1 + 2j, -1 + 1j]), deg=True)
    )
    assert phase_param["unit"] == "deg"


def test__complex_to_real_preparser_complex_setpoint() -> None:

    data_in: List[List[DSPlotData]] = [
        [
            {
                "data": np.array([0 + 0j, 1 + 2j, -1 + 1j]),
                "name": "signal",
                "label": "complex signal",
                "unit": "Ohm",
                "shape": None,
            },
            {
                "data": np.array([0, 1, 2]),
                "name": "voltage",
                "label": "measured voltage",
                "unit": "V",
                "shape": None,
            },
        ]
    ]

    data_out = _complex_to_real_preparser(data_in, conversion='real_and_imag')

    assert np.shape(np.array(data_out)) == (1, 3)
    assert data_out[0][-1] == data_in[0][-1]

    real_param = data_out[0][0]
    assert real_param["name"] == "signal_real"
    assert real_param["label"] == "complex signal [real]"
    assert all(real_param["data"] == np.array([0, 1, -1]))
    assert real_param["unit"] == "Ohm"

    imag_param = data_out[0][1]
    assert imag_param['name'] == 'signal_imag'
    assert imag_param['label'] == 'complex signal [imag]'
    assert all(imag_param['data'] == np.array([0, 2, 1]))
    assert imag_param['unit'] == 'Ohm'

    measured_param = data_out[0][2]
    assert measured_param['name'] == 'voltage'
    assert measured_param['label'] == 'measured voltage'
    assert measured_param['unit'] == 'V'
    assert all(measured_param['data'] == np.array([0, 1, 2]))



from typing import Optional, Type
from types import TracebackType, FrameType
import signal
import logging
import threading

log = logging.getLogger(__name__)


class DelayedKeyboardInterrupt:
    """
    A context manager to wrap a piece of code to ensure that a
    KeyboardInterrupt is not triggered by a SIGINT during the execution of
    this context. A second SIGINT will trigger the KeyboardInterrupt
    immediately.

    Inspired by https://stackoverflow.com/questions/842557/how-to-prevent-a-block-of-code-from-being-interrupted-by-keyboardinterrupt-in-py
    """
    signal_received = None
    old_handler = None

    def __enter__(self) -> None:
        is_main_thread = threading.current_thread() is threading.main_thread()
        is_default_sig_handler = (signal.getsignal(signal.SIGINT)
                                  is signal.default_int_handler)
        if is_default_sig_handler and is_main_thread:
            self.old_handler = signal.signal(signal.SIGINT, self.handler)
        elif is_default_sig_handler:
            log.debug("Not on main thread cannot intercept interrupts")

    def handler(self, sig: signal.Signals, frame: FrameType) -> None:
        self.signal_received = (sig, frame)
        print("Received SIGINT, Will interrupt at first suitable time. "
              "Send second SIGINT to interrupt immediately.")
        # now that we have gotten one SIGINT make the signal
        # trigger a keyboard interrupt normally
        signal.signal(signal.SIGINT, self.forceful_handler)
        log.info('SIGINT received. Delaying KeyboardInterrupt.')

    @staticmethod
    def forceful_handler(sig: signal.Signals, frame: FrameType) -> None:
        print("Second SIGINT received. Triggering "
              "KeyboardInterrupt immediately.")
        log.info('Second SIGINT received. Triggering '
                 'KeyboardInterrupt immediately.')
        signal.default_int_handler(sig, frame)

    def __exit__(self, exception_type: Optional[Type[BaseException]],
                 value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        if self.old_handler is not None:
            signal.signal(signal.SIGINT, self.old_handler)
        if self.signal_received is not None:
            if (self.old_handler is not None and
                    not isinstance(self.old_handler, int)):
                self.old_handler(*self.signal_received)



from typing import Tuple

from qcodes.instrument.base import AbstractInstrument


class KeysightErrorQueueMixin(AbstractInstrument):
    """
    Mixin class for visa instruments that happen to implement an error queue.

    It is meant to work ONLY with drivers for Keysight
    instruments (which inherit from VisaInstrument class).
    """

    def error(self) -> Tuple[int, str]:
        """
        Return the first error message in the queue. It also clears it from
        the error queue.

        Up to 20 errors can be stored in the instrument's error queue.
        Error retrieval is first-in-first-out (FIFO).

        If more than 20 errors have occurred, the most recent error stored
        in the queue is replaced with -350,"Queue overflow". No additional
        errors are stored until you remove errors from the queue. If no
        errors have occurred when you read the error queue, the instrument
        responds with +0,"No error".

        Returns:
            The error code and the error message.
        """
        rawmssg = self.ask('SYSTem:ERRor?')
        code = int(rawmssg.split(',')[0])
        mssg = rawmssg.split(',')[1].strip().replace('"', '')

        return code, mssg

    def flush_error_queue(self, verbose: bool=True) -> None:
        """
        Clear the instrument error queue, and prints it.

        Args:
            verbose: If true, the error messages are printed.
                Default: True.
        """

        self.log.debug('Flushing error queue...')

        err_code, err_message = self.error()
        self.log.debug(f'    {err_code}, {err_message}')
        if verbose:
            print(err_code, err_message)

        while err_code != 0:
            err_code, err_message = self.error()
            self.log.debug(f'    {err_code}, {err_message}')
            if verbose:
                print(err_code, err_message)

        self.log.debug('...flushing complete')



import re
import json
from datetime import datetime, timedelta
import random
from typing import List

import pytest
import hypothesis.strategies as hst
from hypothesis import given, settings


from qcodes.dataset.linked_datasets.links import (
    Link, link_to_str, str_to_link, str_to_links, links_to_str)
from qcodes.dataset.guids import generate_guid


def generate_some_links(N: int) -> List[Link]:
    """
    Generate N links with the same head
    """
    def _timestamp() -> int:
        """
        return a random timestamp that is approximately
        one day in the past.
        """
        timestamp = datetime.now() - timedelta(days=1,
                                               seconds=random.randint(1, 1000))
        return int(round(timestamp.timestamp()*1000))


    known_types = ("fit", "analysis", "step")
    known_descs = ("A second-order fit",
                   "Manual analysis (see notebook)",
                   "Step 3 in the characterisation")

    head_guid = generate_guid(_timestamp())
    head_guids = [head_guid]*N
    tail_guids = [generate_guid(_timestamp()) for _ in range(N)]
    edge_types = [known_types[i % len(known_types)] for i in range(N)]
    descriptions = [known_descs[i % len(known_descs)] for i in range(N)]

    zipattrs = zip(head_guids, tail_guids, edge_types, descriptions)

    links = [Link(hg, tg, n, d) for hg, tg, n, d in zipattrs]

    return links


def test_link_construction_passes():
    head_guid = generate_guid()
    tail_guid = generate_guid()
    edge_type = "fit"
    description = "We did a second order fit with math"

    link = Link(head_guid, tail_guid, edge_type)

    assert link.head == head_guid
    assert link.tail == tail_guid
    assert link.edge_type == edge_type
    assert link.description == ""

    link = Link(head_guid, tail_guid, edge_type, description)

    assert link.description == description


@settings(max_examples=20)
@given(not_guid=hst.text())
def test_link_construction_raises(not_guid):
    head_guid = generate_guid()
    tail_guid = generate_guid()
    edge_type = "fit"

    match = re.escape(
        f'The guid given for head is not a valid guid. Received '
        f'{not_guid}.')
    with pytest.raises(ValueError, match=match):
        Link(not_guid, tail_guid, edge_type)

    match = re.escape(
        f'The guid given for tail is not a valid guid. Received '
        f'{not_guid}')
    with pytest.raises(ValueError, match=match):
        Link(head_guid, not_guid, edge_type)


def test_link_to_str():
    head_guid = generate_guid()
    tail_guid = generate_guid()
    edge_type = "test"
    description = "used in test_link_to_str"

    link = Link(head_guid, tail_guid, edge_type, description)

    expected_str = json.dumps({"head": head_guid,
                               "tail": tail_guid,
                               "edge_type": edge_type,
                               "description": description})

    assert link_to_str(link) == expected_str


def test_str_to_link():
    head_guid = generate_guid()
    tail_guid = generate_guid()
    edge_type = "test"
    description = "used in test_str_to_link"


    lstr = json.dumps({"head": head_guid,
                       "tail": tail_guid,
                       "edge_type": edge_type,
                       "description": description})

    expected_link = Link(head_guid, tail_guid, edge_type, description)

    assert str_to_link(lstr) == expected_link


def test_link_to_string_and_back():
    head_guid = generate_guid()
    tail_guid = generate_guid()
    edge_type = "analysis"
    description = "hyper-spectral quantum blockchain ML"

    link = Link(head_guid, tail_guid, edge_type, description)

    lstr = link_to_str(link)
    newlink = str_to_link(lstr)

    assert newlink == link


@settings(max_examples=5, deadline=1200)
@given(N=hst.integers(min_value=0, max_value=25))
def test_links_to_str_and_back(N):
    links = generate_some_links(N)

    new_links = str_to_links(links_to_str(links))

    assert new_links == links



import numpy as np
import pytest
from qcodes.utils.helpers import make_sweep


def test_good_calls():
    swp = make_sweep(1, 3, num=6)
    assert swp == [1, 1.4, 1.8, 2.2, 2.6, 3]

    swp = make_sweep(1, 3, step=0.5)
    assert swp == [1, 1.5, 2, 2.5, 3]

    # with step, test a lot of combinations with weird fractions
    # to make sure we don't fail on a rounding error
    for r in np.linspace(1, 4, 15):
        for steps in range(5, 55, 6):
            step = r / steps
            swp = make_sweep(1, 1 + r, step=step)
            assert len(swp) == steps + 1
            assert swp[0] == 1
            assert swp[-1] == 1 + r


def test_bad_calls():
    with pytest.raises(AttributeError):
        make_sweep(1, 3, num=3, step=1)

    with pytest.raises(ValueError):
        make_sweep(1, 3)

    # this first one should succeed
    make_sweep(1, 3, step=1)
    # but if we change step slightly (more than the tolerance of
    # 1e-10 steps) it will fail.
    with pytest.raises(ValueError):
        make_sweep(1, 3, step=1.00000001)
    with pytest.raises(ValueError):
        make_sweep(1, 3, step=0.99999999)
